
import morpholib as morpho
import morpholib.anim
from morpholib.tools.basics import *

import pyglet as pg
pyglet = pg
from cmath import exp

# Returns a Frame figure that defines axes over the given view.
#
# ARGUMENTS
# xwidth = x-axis thickness (in pixels). Default: 5
# ywidth = y-axis thickness (in pixels). Default: 5
# xcolor = x-axis color (RGB list). Default: [0,0,0] (black)
# ycolor = y-axis color (RGB list). Default: [0,0,0] (black)
# xalpha = x-axis opacity. Default: 1 (opaque)
# yalpha = y-axis opacity. Default: 1 (opaque)
def Axes(view, xwidth=5, ywidth=5,
    xcolor=(0,0,0), ycolor=(0,0,0), xalpha=1, yalpha=1):
    xAxis = morpho.grid.Path(view[:2])
    xAxis.static = True
    xAxis.width = xwidth
    xAxis.color = list(xcolor)
    xAxis.alpha = xalpha

    yAxis = morpho.grid.Path([view[2]*1j, view[3]*1j])
    yAxis.static = True
    yAxis.width = ywidth
    yAxis.color = list(ycolor)
    yAxis.alpha = yalpha

    return morpho.anim.Frame([xAxis, yAxis])

# Returns a path which is the graph of a real function.
#
# ARGUMENTS
# f = Real to real python function (e.g. lambda x: x**2)
# a,b = Domain interval [a,b]
# steps = Number of line segments in path. Default: 50
# width = Path thickness (in pixels). Default: 3
# color = Path color (RGB list). Default: [1,1,1] (white)
# alpha = Path opacity. Default: 1 (opaque)
def realgraph(f, a, b, steps=50, width=3, color=(1,1,1), alpha=1):
    line = morpho.grid.line(a, b, steps)
    line.width = width
    line.color = list(color)
    line.alpha = alpha

    transform = lambda s: s.real + f(s.real)*1j
    graph = line.fimage(transform)

    return graph



def revolution():
    pass



'''
Various helper functions for dealing with cubic Bezier curves.
'''

# import math, cmath
# import numpy as np
from morpholib.base import lerp0, bezierInterp

# Splits the cubic Bezier curve described by p0...p3 at the
# parameter value t (in interval [0,1]), and returns the control
# points describing the first sub-Bezier curve: the part of the
# curve for lower parameter values than t.
def bezierFirstSlice(p0, p1, p2, p3, t):
    p01 = lerp0(p0, p1, t)
    p12 = lerp0(p1, p2, t)
    p012 = lerp0(p01, p12, t)
    p0123 = bezierInterp(p0, p1, p2, p3, t)

    return (p0, p01, p012, p0123)

# Splits the cubic Bezier curve described by p0...p3 at the
# parameter value t (in interval [0,1]), and returns the control
# points describing the last sub-Bezier curve: the part of the
# curve for higher parameter values than t.
def bezierLastSlice(p0, p1, p2, p3, t):
    return bezierFirstSlice(p3, p2, p1, p0, 1-t)[::-1]
bezierSecondSlice = bezierLastSlice

# Splits the cubic Bezier curve described by p0...p3 at the
# parameter value t (in interval [0,1]), and returns two 4-tuples
# of control points describing the two sub-Bezier curves on either
# side.
def splitBezier(p0, p1, p2, p3, t):
    slice1 = bezierFirstSlice(p0, p1, p2, p3, t)
    slice2 = bezierLastSlice(p0, p1, p2, p3, t)
    return slice1, slice2



'''
Contains some functions useful for making 1080p videos.
'''

import morpholib as morpho
from morpholib.tools.basics import *

import math

ratio = 1920/1080
ratioXY = ratio
ratioYX = 1/ratio
std_view = (-10*ratio, 10*ratio, -10, 10)

def view169():
    return list(std_view)

# Returns a standard video Animation object.
# Meaning an animation with window shape 1920 x 1080, fullscreen.
# A layer list can optionally be supplied. So can background color and opacity.
# Default background and opacity are [1,1,1] (white) and 1 (opaque)
def standardAnimation(layers=None, background=(1,1,1), alpha=1):
    # if layers is None:
    #     layers = morpho.Actor(morpho.Figure)
    if type(layers) not in (tuple, list) and isinstance(layers, morpho.Actor):
        layers = morpho.anim.Layer(layers, view=std_view)
    elif isinstance(layers, morpho.Figure):
        layers = morpho.anim.Layer(morpho.Actor(layers), view=std_view)
    elif layers is None:
        layers = morpho.Layer(view=std_view)

    mation = morpho.Animation(layers)
    mation.windowShape = (1920, 1080)
    mation.fullscreen = True
    mation.resizable = False
    # mation.firstIndex = 0
    mation.background = background[:3]
    mation.alpha = alpha

    # if BGgrid:
    #     nhorz = math.ceil(std_view[3])-math.floor(std_view[2])+1
    #     nvert = math.ceil(std_view[1])-math.floor(std_view[0])+1
    #     hcolor = morpho.tools.color.rgbNormalize(0x23, 0xb5, 0x83)
    #     vcolor = (0,0,1)
    #     grid = morpho.grid.standardGrid(
    #         view=std_view,
    #         nhorz=nhorz, nvert=nvert,
    #         hres=0.05, vres=0.05,
    #         hcolor=hcolor, vcolor=vcolor, alpha=0.5,
    #         hmidColor=morpho.tools.color.alphaOverlay(mation.background[:3], hcolor, 0.5),
    #         vmidColor=morpho.tools.color.alphaOverlay(mation.background[:3], vcolor, 0.5),
    #         BGgrid=False, axesColor=(0,0,0)
    #         )
    #     grid = grid.fimage(lambda s: ((nvert-1)/2)/std_view[1]*s.real + 1j*s.imag)
    #     grid.static = True
    #     grid.delay = oo
    #     grid = morpho.Actor(grid)

    #     layer = morpho.Layer(grid, std_view)
    #     mation.layers.insert(0, layer)
    #     mation.locaterLayer = 0

    return mation


# Essentially same as standardAnimation(), but intended for SpaceLayers.
def standardSpaceAnimation(layers=None, background=(1,1,1), alpha=1):
    # if layers is None:
    #     layers = morpho.Actor(morpho.Figure)
    if type(layers) not in (tuple, list) and isinstance(layers, morpho.Actor):
        layers = morpho.anim.SpaceLayer(layers, view=std_view)
    elif isinstance(layers, morpho.Figure):
        layers = morpho.anim.SpaceLayer(morpho.Actor(layers), view=std_view)
    elif layers is None:
        layers = morpho.SpaceLayer(view=std_view)

    mation = morpho.Animation(layers)
    mation.windowShape = (1920, 1080)
    mation.fullscreen = True
    mation.resizable = False
    # mation.firstIndex = 0
    mation.background = background[:3]
    mation.alpha = alpha

    return mation


# Sets up a standard 3D animation viewing the first octant.
#
# ARGUMENTS
# t_rot = Camera rotation duration (in frames). Default: 300
# initOrient = Initial camera orientation before tilting and spinning.
#              Default: np.eye(3)
# spin = Angular distance the camera will revolve (in rad).
#        Default: -80*pi/180 (80 degs clockwise)
# tilt = Amount camera is tilted down from viewing the xy-plane from above.
#        Default: 70*pi/180 (70 degs)
def setupSpace(t_rot=300, initOrient=np.identity(3), spin=-80*tau/360,
    tilt=70*tau/360):

    xAxis = morpho.grid.SpaceArrow(0, 7.5)
    xAxis.color = [0, 0.5, 0]
    xAxis.width = 7
    xAxis.headSize = 20

    yAxis = xAxis.copy()
    yAxis.head = 7.5j
    yAxis.color = [0, 0, 0.7]

    zAxis = xAxis.copy()
    zAxis.head = [0,0,4.5]
    zAxis.color = [0.7,0,0]

    xLabel = morpho.text.SpaceText(
        text="x", pos=xAxis.head + [0, -0.25, 0.25],
        size=55, italic=True,
        anchor_x=0, anchor_y=0,
        color=[0,0,0]
        )
    yLabel = morpho.text.SpaceText(
        text="y", pos=yAxis.head + [-0.25, 0, 0.25],
        size=55, italic=True,
        anchor_x=0, anchor_y=0,
        color=[0,0,0]
        )
    zLabel = morpho.text.SpaceText(
        text="z", pos=zAxis.head + [0,0,0.25],
        size=55, italic=True,
        anchor_x=0, anchor_y=0,
        color=[0,0,0]
        )

    orient = initOrient
    theta0 = -pi/2 - 5*pi/180
    orient = morpho.matrix.rotation([0,0,1], theta0) @ orient
    orient = morpho.matrix.rotation([1,0,0], -tilt) @ orient
    focus = [3.5, 3.5, 2.5]
    # focus = 0
    layer = morpho.SpaceLayer(orient=orient, focus=focus)
    layer.camera.time(0).view = list(std_view)

    # orient = morpho.matrix.rotation([0,0,1], -pi + 5*pi/180)
    orient = initOrient
    orient = morpho.matrix.rotation([0,0,1], theta0+spin) @ orient
    orient = morpho.matrix.rotation([1,0,0], -tilt) @ orient
    layer.camera.first().transition = morpho.transition.uniform
    layer.camera.first().centerAt(3.5 + 3.5j)
    layer.camera.first().zoomIn(2.5)
    layer.camera.newkey(t_rot)
    layer.camera.last().orient = orient

    # Grid
    grid = morpho.grid.wireframe(
        view=[0,7, 0,7],
        dx=1, dy=1,
        hnodes=2, vnodes=2,
        hcolor=morpho.color.colormap["green"], vcolor=[0,0,1],
        axes=False, optimize=True
        )
    # grid.figures.pop(14)
    # grid.figures.pop(21)
    # grid.figures = morpho.grid.optimizePathList(grid.figures)
    grid = morpho.Frame(grid.figures)
    layer.merge(grid)

    # Merge axes and labels
    layer.merge(xAxis)
    layer.merge(yAxis)
    layer.merge(zAxis)
    layer.merge(xLabel)
    layer.merge(yLabel)
    layer.merge(zLabel)

    # mation = standardAnimation(background=(1,1,1,1))
    # mation.layers[0] = layer

    mation = standardSpaceAnimation(layer)

    return mation


# Sets up an alternative 3D animation object centered around the z-axis.
#
# ARGUMENTS
# t_rot = Camera rotation duration (in frames). Default: 300
# initOrient = Initial camera orientation before tilting and spinning.
#              Default: np.eye(3)
# spin = Angular distance the camera will revolve (in rad).
#        Default: -80*pi/180 (80 degs clockwise)
# tilt = Amount camera is tilted down from viewing the xy-plane from above.
#        Default: 70*pi/180 (70 degs)
def setupSpaceAlt(t_rot=300, initOrient=np.identity(3), spin=-80*tau/360,
    tilt=70*tau/360):

    xAxis = morpho.grid.SpaceArrow(0, 4.5)
    xAxis.color = [0, 0.5, 0]
    xAxis.width = 7
    xAxis.headSize = 20

    yAxis = xAxis.copy()
    yAxis.head = 4.5j
    yAxis.color = [0, 0, 0.7]

    # zAxis = xAxis.copy()
    # zAxis.head = [0,0,4.5]
    # zAxis.color = [0.7,0,0]

    xLabel = morpho.text.SpaceText(
        text="x", pos=xAxis.head + [0.4,0-0.25,0],
        size=55, italic=True,
        anchor_x=0, anchor_y=0,
        color=[0,0,0]
        )
    yLabel = morpho.text.SpaceText(
        text="y", pos=yAxis.head + [-0.25,0.4,0.1],
        size=55, italic=True,
        anchor_x=0, anchor_y=0,
        color=[0,0,0]
        )
    # zLabel = morpho.text.SpaceText(
    #     text="z", pos=zAxis.head + [0,0,0.25],
    #     size=55, italic=True,
    #     anchor_x=0, anchor_y=0,
    #     color=[0,0,0]
    #     )

    orient = initOrient
    theta0 = -pi/2 - 5*pi/180
    orient = morpho.matrix.rotation([0,0,1], theta0) @ orient
    orient = morpho.matrix.rotation([1,0,0], -tilt) @ orient
    focus = [0, 0, 2]
    # focus = 0
    layer = morpho.SpaceLayer(orient=orient, focus=focus)
    layer.camera.time(0).view = list(std_view)

    orient = initOrient
    theta1 = theta0 + spin
    orient = morpho.matrix.rotation([0,0,1], theta1) @ orient
    orient = morpho.matrix.rotation([1,0,0], -tilt) @ orient
    layer.camera.first().transition = morpho.transition.uniform
    # layer.camera.first().centerAt(3.5 + 3.5j)
    layer.camera.first().zoomIn(2.5)
    layer.camera.newkey(t_rot)
    layer.camera.last().orient = orient

    # Grid
    grid = morpho.grid.wireframe(
        view=[-4,4, -4,4],
        dx=1, dy=1,
        hnodes=2, vnodes=2,
        hcolor=morpho.color.colormap["green"], vcolor=[0,0,1],
        axes=False
        )
    grid = morpho.Frame(grid.figures)
    layer.merge(grid)

    layer.merge(xAxis)
    layer.merge(yAxis)
    # layer.merge(zAxis)
    layer.merge(xLabel)
    layer.merge(yLabel)
    # layer.merge(zLabel)

    # mation = standardAnimation(background=(1,1,1,1))
    # mation.layers[0] = layer

    mation = standardSpaceAnimation(layer)

    return mation



import math, cmath
import numpy as np
from bisect import bisect_right, bisect_left

### CONSTANTS ###
pi = math.pi
tau = 2*pi
oo = inf = float("inf")
nan = float("nan")


### FUNCTIONS ###

isbadnum_old = lambda x: math.isnan(abs(x)*0)

# Detect infinite or nan (real or complex)
def isbadnum(x):
    return cmath.isnan(x) or cmath.isinf(x)

# isbadnum() but for np.arrays. Returns True if and only if
# nan or inf is found ANYWHERE in the array.
# Actually, this function should also work on scalars (both numpy
# and python builtins), so it can totally replace isbadnum().
# However, it might run slower than isbadnum() on python scalars,
# so I'm keeping the original isbadnum() around anyway.
def isbadarray(x):
    return (np.any(np.isnan(x)) or np.any(np.isinf(x))).tolist()

# Conversion factors between degrees and radians.
deg = tau/360
rad = 1/deg

# Real cotangent function
PI_OVER_2 = pi/2
def cot(theta):
    return math.tan(PI_OVER_2 - theta)

# Signum function. Returns x/abs(x) unless x = 0, in which case,
# return 0
def sgn(x):
    if x > 0:
        return 1
    elif x < 0:
        return -1
    else:
        return 0

# Computes the mean of a vector-like thing.
def mean(a):
    if len(a) == 0:
        raise IndexError("Can't take mean of an empty list-like object!")
    return sum(a)/len(a)

# Like round(), but truncates instead of rounding at the decimal digit you
# specify
def truncate(num, ndigits):
    decshift = 10**ndigits
    return int(num*decshift)/decshift

# Used in the spiral tween method.
# Computes the correct amount to shift an angle th1 so that it
# becomes th2 in the shortest possible path
# i.e. a path that does not traverse more than pi radians
# The value returned is called "dth" and should be used in
# expressions like this: th(t) = th1 + t*dth
# However, before using the above expression, make sure th1 and th2
# are modded 2pi.
# (Actually, maybe it's okay if they're not?)
def argShift(th1, th2):
    th1 = th1 % tau
    th2 = th2 % tau

    dth = th2 - th1
    if abs(dth) > pi + 1.189e-12:
        dth = dth - math.copysign(tau, dth)
    return dth

# Identical to argShift(), but works on np.arrays
def argShiftArray(th1, th2):
    th1 = th1 % tau
    th2 = th2 % tau

    dth = th2 - th1

    flagset = abs(dth) > pi + 1.189e-12
    subset = dth[flagset]
    dth[flagset] = subset - np.copysign(tau, subset)

    return dth

# Given two points in the complex plane and the angle (in radians)
# of the circular arc that is supposed to go between them,
# returns the center point of the arc (as a complex number).
# Note that the angle is signed. Positive means the arc
# travels CCW from p to q.
# Also note that the angle cannot be 0 or a multiple of 2pi.
def arcCenter(p, q, angle):
    if angle % tau == 0:
        raise ValueError("Angle is 0 or a multiple of 2pi. Center point is undefined!")

    m = (p+q)/2  # midpoint
    return m + 1j*(m-p)*cot(angle/2)  # center point

# Array version of arcCenter()
def arcCenterArray(p, q, angle):
    if np.any(angle % tau == 0):
        raise ValueError("Angle is 0 or a multiple of 2pi. Center point is undefined!")

    m = (p+q)/2  # midpoint
    halfangle = angle/2
    return m + 1j*(m-p)*(np.cos(halfangle)/np.sin(halfangle))

# Converts a 2D tuple/list into a complex number
def vect2complex(v):
    return v[0] + 1j*v[1]

# Given a sorted list of numbers a and value x,
# returns the highest index i such that a[i] <= x.
# If all the numbers in a are larger than x, it returns -1.
def listfloor(a, x):
    return bisect_right(a,x)-1

# Given a sorted list of numbers a and value x,
# returns the lowest index i such that a[i] >= x.
# If all the numbers in a are less than x, it returns len(a).
def listceil(a, x):
    if x in a:
        return a.index(x)
    else:
        return listfloor(a,x) + 1

# If x is a float that is really an integer, returns int(x).
# Otherwise, just returns back x unchanged.
def flattenFloat(x):
    if type(x) is float and x == int(x):
        x = int(x)
    return x

# Returns the functional composition of all the functions provided.
# e.g. Given compose(f,g,h), returns f o g o h = f(g(h(*)))
def compose(*funcs):
    if len(funcs) == 0:
        raise ValueError("No functions to compose!")
    def composition(*args, **kwargs):
        value = funcs[-1](*args, **kwargs)
        for n in range(len(funcs)-2,-1,-1):
            value = funcs[n](value)
        return value

    return composition

# Returns the functional composition of f with g: (f o g)
def compose2(f,g):
    return lambda *args, **kwargs: f(g(*args, **kwargs))

# Given a standard viewbox [xmin,xmax, ymin,ymax],
# returns the corners of the box as a list of complex numbers.
# Useful for passing into a Path or Polygon figure's seq or vertices.
# By default, it does so starting at the northwest corner and
# going counter-clockwise, but this can be modified by altering
# the "initCorner" and "CCW" parameters. By default:
# initCorner = "NW"
# CCW = True
def boxCorners(box, initCorner="NW", CCW=True):
    initCorner = initCorner.upper()
    dirs = ["NW", "SW", "SE", "NE"]
    if initCorner not in dirs:
        raise ValueError('initCorner must be "NW", "SW", "SE", or "NE".')

    left = box[0]
    right = box[1]
    bottom = box[2]
    top = box[3]

    corners = [left+1j*top, left+1j*bottom, right+1j*bottom, right+1j*top]

    # Reverse order if done clockwise
    if not CCW:
        corners = corners[::-1]
        dirs = dirs[::-1]

    # Find the index of the starting corner
    i = dirs.index(initCorner)

    # Order the corners according to the starting corner and direction
    corners = corners[i:] + corners[:i]

    return corners




import morpholib as morpho
import morpholib.anim, morpholib.grid
from morpholib.tools.basics import *

import cairo
cr = cairo

import numpy as np

# Setup a dummy context for use in computing certain text values
# outside of draw time (e.g. width and height)
mation = morpho.Animation()
mation.setupContext()
ctx0 = mation.context
del mation

### CLASSES ###

'''
anchor_x and anchor_y should be interpreted as defining where the
origin for the text is. (0,0) corresponds to centered,
(1,1) corresponds to the upper-right corner,
(-1, -1) corresponds to the lower-left corner.
'''

# Basic text figure. Displays uniformly formatted text on screen.
#
# TWEENABLES
# pos = Position of text (complex number). Default: 0
# size = Text size (in pixels? Not totally sure). Default: 64
# anchor_x = Horizontal alignment parameter.
#            -1 = left-aligned, 0 = center-aligned, 1 = right-aligned.
#            Default: 0 (center-aligned)
# anchor_y = Vertical alignment parameter.
#            -1 = bottom-aligned, 0 = center-aligned, 1 = top-aligned.
#            Default: 0 (center-aligned)
# color = Text color (RGB list). Default: (1,1,1) (white)
# alpha = Opacity. Default: 1 (opaque)
# background = Background box color. Default: [1,1,1] (white)
# backAlpha = Background box opacity. Default: 0 (transparent)
# backPad = Background box padding (physical units). Default: 0
# rotation = Rotation in radians. Default: 0
# prescale_x, prescale_y = Scale factors applied to the text
#             BEFORE the rotation tweenable is applied.
#             Mainly for internal use.
#
# OTHER ATTRIBUTES
# text = Text to display (string). Default: "" (empty string)
# font = Font to use. Default: "Times New Roman"
# bold = Boolean indicating whether to bold. Default: False
# italic = Boolean indicating whether to use italics. Default: False
class Text(morpho.Figure):
    def __init__(self, text="", pos=complex(0),
        size=64, font="Times New Roman",
        bold=False, italic=False,
        anchor_x=0, anchor_y=0,
        color=(1,1,1), alpha=1,
        background=(1,1,1), backAlpha=0, backPad=0):

        super().__init__()

        # Handle Text figure derivative inputs for text.
        # Just extract the "text" attribute.
        if isinstance(text, Text) or isinstance(text, MultiText) \
            or isinstance(text, SpaceMultiText):
            text = text.text

        if type(color) is tuple:
            color = list(color)
        elif type(color) is not list:
            raise TypeError("Unsupported color input")

        # Take last three coords of color.
        color = list(color[:3])

        # Tack on zeros if len(color) < 4
        if len(color) < 3:
            color.extend([0]*(4-len(color)))

        # Turn position into complex
        if type(pos) in (list, tuple):
            pos = pos[0] + 1j*pos[1]

        # Create tweenables
        pos = morpho.Tweenable("pos", pos, tags=["complex"])
        size = morpho.Tweenable("size", size, tags=["size"])
        anchor_x = morpho.Tweenable("anchor_x", anchor_x, tags=["scalar"])
        anchor_y = morpho.Tweenable("anchor_y", anchor_y, tags=["scalar"])
        _transform = morpho.Tweenable("_transform", np.identity(2), tags=["nparray"])
        color = morpho.Tweenable("color", color, tags=["color"])
        alpha = morpho.Tweenable("alpha", alpha, tags=["scalar"])
        background = morpho.Tweenable("background", background, tags=["color"])
        backAlpha = morpho.Tweenable("backAlpha", backAlpha, tags=["scalar"])
        backPad = morpho.Tweenable("backPad", backPad, tags=["scalar"])
        # CCW rotation in radians
        rotation = morpho.Tweenable("rotation", 0, tags=["scalar"])

        # These are the pre-transformation scale factors. These get
        # applied to the text BEFORE rotation and transform do.
        # This is mainly for use in the Multi decorator, so that
        # tweening between two texts with differing rotation parameters
        # works correctly.
        prescale_x = morpho.Tweenable("prescale_x", 1, tags=["scalar"])
        prescale_y = morpho.Tweenable("prescale_y", 1, tags=["scalar"])

        self.update([pos, size, anchor_x, anchor_y, _transform,
            color, alpha, background, backAlpha, backPad, rotation,
            prescale_x, prescale_y])

        # Other attributes
        self.text = text
        self.font = font
        self.bold = bold
        self.italic = italic

    @property
    def transform(self):
        return self._transform

    @transform.setter
    def transform(self, value):
        self._transform = morpho.matrix.array(value)


    def copy(self):
        # Do a standard figure copy first
        # new = morpho.Figure.copy(self)
        new = super().copy()

        # Copy the non-tweenable attributes
        new.text = self.text
        new.font = self.font
        new.bold = self.bold
        new.italic = self.italic

        return new

    # Returns the dimensions (in pixels) of the text as a pair
    # (textwidth, textheight).
    # Note: This ignores the transform attribute.
    def pixelDimensions(self):
        ctx0.select_font_face(
            self.font,
            cr.FONT_SLANT_ITALIC if self.italic else cr.FONT_SLANT_NORMAL,
            cr.FONT_WEIGHT_BOLD if self.bold else cr.FONT_WEIGHT_NORMAL
            )
        ctx0.set_font_size(self.size)
        ctx0.set_source_rgba(*self.color, self.alpha)

        # Compute alignment parameters
        anchor_x = (self.anchor_x + 1)/2
        anchor_y = (self.anchor_y + 1)/2
        xDummy, yDummy, textWidth, textHeight, dx, dy = ctx0.text_extents(self.text)

        return (textWidth, textHeight)

    # Given viewbox and cairo context (or windowShape tuple),
    # returns tuple (width, height) representing the text's physical
    # width and height.
    # Note this ignores the "transform" tweenable.
    def dimensions(self, view, ctx):
        if isinstance(view, morpho.anim.Camera):
            view = view.view

        WIDTH, HEIGHT = self.pixelDimensions()
        width = morpho.physicalWidth(WIDTH, view, ctx)
        height = morpho.physicalHeight(HEIGHT, view, ctx)

        return (width, height)

    # Returns bounding box of the text in physical units.
    # Mainly of use internally to draw the background box.
    # Note: Ignores rotation and prescale factors.
    def box(self, view, ctx, pad=0):
        width, height = self.dimensions(view, ctx)
        # Modify by prescale factors
        width *= self.prescale_x
        height *= self.prescale_y
        align_x, align_y = self.anchor_x, self.anchor_y
        a = self.pos.real - width/2*(align_x + 1)
        b = a + width
        c = self.pos.imag - height/2*(align_y + 1)
        d = c + height

        return [a-pad, b+pad, c-pad, d+pad]

    # Same as box(), but the coordinates are relative to
    # the text's position.
    def relbox(self, view, ctx, pad=0):
        width, height = self.dimensions(view, ctx)
        # Modify by prescale factors
        width *= self.prescale_x
        height *= self.prescale_y
        align_x, align_y = self.anchor_x, self.anchor_y
        a = -width/2*(align_x + 1)
        b = a + width
        c = -height/2*(align_y + 1)
        d = c + height

        return [a-pad, b+pad, c-pad, d+pad]


    # Returns the width of the text in pixels.
    def pixelWidth(self):
        return self.pixelDimensions()[0]

    # Returns the height of the text in pixels.
    def pixelHeight(self):
        return self.pixelDimensions()[1]

    # Returns the physical width of the text.
    # Same as mytext.dimensions(view, ctx)[0]
    def width(self, view, ctx):
        return self.dimensions(view, ctx)[0]

    # Returns the physical height of the text.
    # Same as mytext.dimensions(view, ctx)[1]
    def height(self, view, ctx):
        return self.dimensions(view, ctx)[1]

    def draw(self, camera, ctx):
        # Do nothing if size less than 1.
        if self.size < 1:
            return

        view = camera.view

        x,y = morpho.anim.screenCoords(self.pos, view, ctx)

        ctx.select_font_face(
            self.font,
            cr.FONT_SLANT_ITALIC if self.italic else cr.FONT_SLANT_NORMAL,
            cr.FONT_WEIGHT_BOLD if self.bold else cr.FONT_WEIGHT_NORMAL
            )
        ctx.set_font_size(self.size)
        ctx.set_source_rgba(*self.color, self.alpha)

        # Compute alignment parameters
        anchor_x = (self.anchor_x + 1)/2
        anchor_y = (self.anchor_y + 1)/2
        xDummy, yDummy, textWidth, textHeight, dx, dy = ctx.text_extents(self.text)

        # # Check if transformation matrix is too close to singular.
        # # Specifically, is the area covered by the transformed text
        # # smaller than a single pixel?
        # if abs(np.linalg.det(self.transform)*textWidth*textHeight) < 1:
        #     return

        # Check if the text has been distorted too thin.
        # Specifically, is the thinnest height of the parallelogram
        # it spans less then a pixel? If so, don't draw!
        mat = self.transform.copy()
        mat[:,0] *= textWidth*self.prescale_x
        mat[:,1] *= textHeight*self.prescale_y
        if morpho.matrix.thinHeight2x2(mat) < 1:
            return

        if self.backAlpha > 0:
            # Construct background rectangle and draw it
            box = self.relbox(view, ctx, pad=self.backPad)
            rect = morpho.grid.rect(box)
            rect.origin = self.pos
            rect.width = 0
            rect.fill = self.background
            rect.alpha = self.backAlpha*self.alpha
            rect.rotation = self.rotation
            rect.transform = self.transform
            rect.draw(camera, ctx)

        ctx.save()

        # ctx.translate(x,-y)
        # ctx.translate(0, 2*y)
        ctx.translate(x,y)

        # Apply transformation matrix if necessary
        if not np.array_equal(self.transform, np.identity(2)):
            # Define cairo matrix
            xx, xy, yx, yy = self.transform.flatten()
            mat = cairo.Matrix(xx, yx, xy, yy)

            # Apply to context
            ctx.transform(mat)

        if (self.rotation % tau) != 0:
            ctx.rotate(self.rotation)

        # Apply pre-transformation scales
        ctx.scale(self.prescale_x, self.prescale_y)

        # Handle alignment
        ctx.translate(-anchor_x*textWidth, -anchor_y*textHeight)

        ctx.scale(1,-1)

        # ctx.move_to(x-anchor_x*textWidth, y+anchor_y*textHeight)
        ctx.move_to(0,0)

        try:
            ctx.show_text(self.text)
        except cairo.Error:
            pass

        ctx.restore()
        ctx.new_path()



# Decorator for tween methods in the MultiText class below.
# Reworks ordinary Text class tween methods so that they work
# in a multifigure setting.
#
# Optionally specify a method called "reverseMethod" which is used
# instead of the main method when the main method would have been
# called "in reverse" by calling textMethod(other, self, 1-t).
# This was originally developed to solve the problem of decorating
# tweenPivot() because it is not symmetric in swapping
# self with other.
def Multi(imageMethod, reverseMethod=None):
    if reverseMethod is None:
        reverseMethod = imageMethod

    def wrapper(self, other, t, *args, **kwargs):

        diff = len(self.figures) - len(other.figures)
        if diff > 0:
            # Temporarily extend the image list of other with copies of
            # other's final image
            extension = []
            for i in range(diff):
                extension.append(other.figures[-1].copy())
            other.figures.extend(extension)
            tw = wrapper(self, other, t)
            # Restore other to its original state
            other.figures = other.figures[:-diff]
            return tw
        elif diff < 0:
            # Temporarily extend the image list of self with copies of
            # self's final image
            extension = []
            for i in range(-diff):
                extension.append(self.figures[-1].copy())
            self.figures.extend(extension)
            tw = wrapper(self, other, t)
            self.figures = self.figures[:diff]
            return tw

        figures = []
        for n in range(len(self.figures)):
            selffig = self.figures[n]
            otherfig = other.figures[n]

            # If both underlying figures have the same text, don't do anything fancy.
            if selffig.text == otherfig.text:
                new = imageMethod(selffig, otherfig, t, *args, **kwargs)
                figures.append(new)
            # Fade out self and fade in other
            else:
                # Compute the scale matrices
                selfWidth, selfHeight = selffig.pixelDimensions()
                otherWidth, otherHeight = otherfig.pixelDimensions()
                forward_scale_x = otherWidth / selfWidth
                forward_scale_y = otherHeight / selfHeight
                backward_scale_x = 1/forward_scale_x
                backward_scale_y = 1/forward_scale_y
                # # Scale matrix for self so that it fits other at t=1
                # fwdMatrix = np.array(
                #     [[forward_scale_x, 0],
                #      [0, forward_scale_y]], dtype=float
                #      )
                # # Scale matrix for other so that it fits self at t=0
                # backMatrix = np.array(
                #     [[1/forward_scale_x, 0],
                #      [0, 1/forward_scale_y]], dtype=float
                #      )

                # if (self.rotation % tau) != 0:
                #     c,s = math.cos(self.rotation), math.sin(self.rotation)
                #     selfRot = np.array([[c,-s],[s,c]], dtype=float)
                #     backMatrix = selfRot @ backMatrix @ selfRot.T
                # if (other.rotation % tau) != 0:
                #     c,s = math.cos(other.rotation), math.sin(other.rotation)
                #     otherRot = np.array([[c,-s],[s,c]], dtype=float)
                #     fwdMatrix = otherRot @ fwdMatrix @ otherRot.T

                selffig1 = otherfig.copy()
                # selffig1.transform = selffig1.transform @ fwdMatrix
                selffig1.prescale_x *= forward_scale_x
                selffig1.prescale_y *= forward_scale_y
                selffig1.alpha = 0

                otherfig0 = selffig.copy()
                # otherfig0.transform = otherfig0.transform @ backMatrix
                otherfig0.prescale_x *= backward_scale_x
                otherfig0.prescale_y *= backward_scale_y
                otherfig0.alpha = 0

                newself = imageMethod(selffig, selffig1, t, *args, **kwargs)
                newother = reverseMethod(otherfig, otherfig0, 1-t, *args, **kwargs)

                figures.append(newself)
                figures.append(newother)

        # Remove temporary extensions
        if diff > 0:
            other.figures = other.figures[:-len(extensions)]
        elif diff < 0:
            self.figures = self.figures[:-len(extensions)]

        tw = type(self)(figures)
        tw.defaultTween = self.defaultTween
        tw.transition = self.transition
        tw.static = self.static
        tw.delay = self.delay
        tw.visible = self.visible

        # The following handling of zdepth seems a little too
        # hard-coded.
        # It assumes you always want to linearly tween zdepth.
        # If you plan on using the Multi decorator more broadly,
        # you should consider reimplementing this.
        tw.zdepth = morpho.numTween(self.zdepth, other.zdepth, t)

        return tw

    return wrapper


# Text class that can support drawing multiple Text figures at once.
# Useful for having one text morph into another text.
#
# See "morpho.graphics.MultiImage" for more info on the basic idea here.
#
# Bottom line: It's just like Text except you can tween between different
# underlying text strings.
class MultiText(morpho.MultiFigure):
    def __init__(self, text="", *args, **kwargs):
        if isinstance(text, str):
            textlist = [Text(text, *args, **kwargs)]
        elif isinstance(text, list) or isinstance(text, tuple):
            textlist = [(Text(item, *args, **kwargs) if isinstance(item, str) else item) for item in text]
        else:
            textlist = [Text(text, *args, **kwargs)]

        # Create frame figure
        super().__init__(textlist)

    @property
    def textlist(self):
        return self.figures

    @textlist.setter
    def textlist(self, value):
        self.figures = value

        # Convert every figure in the list to a Text figure
        # if possible.
        for n in range(len(self.figures)):
            fig = self.figures[n]
            if not isinstance(fig, Text):
                newfig = fig.images[0].copy()
                self.figures[n] = newfig

    ### TWEEN METHODS ###

    tweenLinear = Multi(Text.tweenLinear)
    tweenSpiral = Multi(Text.tweenSpiral)

    @classmethod
    def tweenPivot(cls, angle=tau/2, *args, **kwargs):
        return Multi(Text.tweenPivot(angle, *args, **kwargs),
            reverseMethod=Text.tweenPivot(-angle, *args, **kwargs)
            )




# DEPRECATED. Use "Text" instead.
class Text_old(morpho.Figure):
    def __init__(self, text=None, pos=complex(0),
        size=18, font="Times New Roman",
        bold=False, italic=False,
        anchor_x=-1, anchor_y=0,
        color=(1,1,1), alpha=1,
        physical=True):

        morpho.Figure.__init__(self)

        if text is None:
            text = []
        elif type(text) is str or isinstance(text, Number):
            text = [text]
        elif type(text) is tuple:
            text = list(text)
        elif type(text) is not list:
            raise TypeError("Unsupported text input")

        if type(color) is tuple:
            color = list(color)
        elif type(color) is not list:
            raise TypeError("Unsupported color input")

        # Take last three coords of color.
        color = list(color[:3])

        # Tack on zeros if len(color) < 4
        if len(color) < 3:
            color.extend([0]*(4-len(color)))

        # Turn position into complex
        if type(pos) in (list, tuple):
            pos = pos[0] + 1j*pos[1]

        # Create tweenables
        textlist = morpho.Tweenable("textlist", text, tags=["textlist", "nolinear", "nospiral"])
        pos = morpho.Tweenable("pos", pos, tags=["complex"])
        size = morpho.Tweenable("size", size, tags=["size"])
        _transform = morpho.Tweenable("_transform", np.identity(2), tags=["nparray"])
        color = morpho.Tweenable("color", color, tags=["color"])
        alpha = morpho.Tweenable("alpha", alpha, tags=["scalar"])
        self.update([textlist, pos, size, _transform, color, alpha])

        # Other attributes
        self.anchor_x = anchor_x
        self.anchor_y = anchor_y
        self.font = font
        self.bold = bold
        self.italic = italic
        self.physical = physical

        # self.defaultTween = Text.tweenLinear

    # Allows accessing the zeroth item of the textlist via
    # mytext.text
    @property
    def text(self):
        return self.textlist[0]

    # Allows modifying the zeroth item of the textlist via
    # mytext.text = "haha lolz"
    @text.setter
    def text(self, value):
        self.textlist[0] = value

    @property
    def transform(self):
        return self._transform

    @transform.setter
    def transform(self, value):
        self._transform = morpho.matrix.array(value)


    def copy(self):
        # Do a standard figure copy first
        new = morpho.Figure.copy(self)

        # Do a deep copy of the textlist
        for i in range(len(new.textlist)):
            item = new.textlist[i]
            if type(item) is Number:
                new.textlist[i] = item.copy()

        # Copy the non-tweenable attributes
        new.anchor_x = self.anchor_x
        new.anchor_y = self.anchor_y
        new.font = self.font
        new.bold = self.bold
        new.italic = self.italic
        new.physical = self.physical
        return new

    def draw(self, camera, ctx):
        # Do nothing if size less than 1.
        if self.size < 1:
            return

        view = camera.view

        # Compute literal text to draw
        textlist = []
        for item in self.textlist:
            textlist.append(str(item))

        text = "".join(textlist)

        # If the text is physical, position it in the complex plane,
        # else treat position as window pixel coordinates.
        if self.physical:
            x,y = morpho.anim.screenCoords(self.pos, view, ctx)
        else:
            x,y = self.pos.real, self.pos.imag

        ctx.select_font_face(
            self.font,
            cr.FONT_SLANT_ITALIC if self.italic else cr.FONT_SLANT_NORMAL,
            cr.FONT_WEIGHT_BOLD if self.bold else cr.FONT_WEIGHT_NORMAL
            )
        ctx.set_font_size(self.size)
        ctx.set_source_rgba(*self.color, self.alpha)

        # Compute alignment parameters
        anchor_x = (self.anchor_x + 1)/2
        anchor_y = (self.anchor_y + 1)/2
        xDummy, yDummy, textWidth, textHeight, dx, dy = ctx.text_extents(text)

        # Check if transformation matrix is too close to singular.
        # Specifically, is the area covered by the transformed text
        # smaller than a single pixel?
        if abs(np.linalg.det(self.transform)*textWidth*textHeight) < 1:
            return

        ctx.save()

        # ctx.translate(x,-y)
        # ctx.translate(0, 2*y)
        ctx.translate(x,y)

        # Apply transformation matrix if necessary
        if not np.array_equal(self.transform, np.identity(2)):
            # Define cairo matrix
            xx, xy, yx, yy = self.transform.flatten()
            mat = cairo.Matrix(xx, yx, xy, yy)

            # Apply to context
            ctx.transform(mat)

        # Handle alignment
        ctx.translate(-anchor_x*textWidth, -anchor_y*textHeight)

        ctx.scale(1,-1)

        # ctx.move_to(x-anchor_x*textWidth, y+anchor_y*textHeight)
        ctx.move_to(0,0)

        ctx.show_text(text)
        ctx.restore()
        ctx.new_path()

    ### TWEEN METHODS ###

    @morpho.TweenMethod
    def tweenLinear(self, other, t):
        # Tween all tweenables except the textlist
        txt = morpho.Figure.tweenLinear(self, other, t)
        # Don't tween the position if it is not physical
        if not self.physical:
            txt.pos = self.pos

        # Now handle the textlist
        for i in range(len(txt.textlist)):
            item = txt.textlist[i]
            if type(item) is not Number: continue

            self_Number = self.textlist[i]
            other_Number = other.textlist[i]

            txt.textlist[i] = self_Number.defaultTween(
                self_Number, other_Number, t)

        return txt

    @morpho.TweenMethod
    def tweenSpiral(self, other, t):
        # Tween all tweenables except the textlist
        txt = morpho.Figure.tweenSpiral(self, other, t)
        # Don't tween the position if it is not physical
        if not self.physical:
            txt.pos = self.pos

        # Now handle the textlist
        for i in range(len(txt.textlist)):
            item = txt.textlist[i]
            if type(item) is not Number: continue

            self_Number = self.textlist[i]
            other_Number = other.textlist[i]

            txt.textlist[i] = self_Number.defaultTween(
                self_Number, other_Number, t)

        return txt

    # @morpho.TweenMethod
    # def tweenLinearZoom(self, other, t):
    #     # Tween all tweenables except the textlist
    #     txt = morpho.Figure.tweenLinear(self, other, t)
    #     # Don't tween the position if it is not physical
    #     if not self.physical:
    #         txt.pos = self.pos

    #     # Now handle the textlist
    #     for i in range(len(txt.textlist)):
    #         item = txt.textlist[i]
    #         if type(item) is not Number: continue

    #         a = self.textlist[i].number
    #         b = other.textlist[i].number
    #         item.number = a*(b/a)**t

    #     return txt


# 3D version of the Text class.
#
# TWEENABLES that are not shared with Image
# orient = Orientation of text relative to text position (3x3 np.array).
#          Only used if "orientable" attribute is set to True.
#          Default: np.eye(3) meaning text is oriented flat on xy-plane facing
#          in the positive z direction.
#
# OTHER ATTRIBUTES
# orientable = Boolean specifying whether text should be orientable in 3D space,
#              or just behave like a label always facing the camera. Default: False
class SpaceText(Text):
    def __init__(self, text=None, pos=None,
        size=64, font="Times New Roman",
        bold=False, italic=False,
        anchor_x=0, anchor_y=0,
        color=(1,1,1), alpha=1,
        background=(1,1,1), backAlpha=0, backPad=0):

        super().__init__(text, 0,
            size, font,
            bold, italic,
            anchor_x, anchor_y,
            color[:], alpha,
            background, backAlpha, backPad
            )

        if pos is None:
            pos = np.zeros(3)

        # Redefine pos tweenable to be 3D.
        _pos = morpho.Tweenable("_pos", morpho.matrix.array(pos), tags=["nparray", "fimage"])
        self._state.pop("pos")
        self._state["_pos"] = _pos
        _orient = morpho.Tweenable("_orient", np.identity(3), tags=["nparray", "orient"])
        self._state["_orient"] = _orient

        self.orientable = False

        # # Redefine pos tweenable to be 3D.
        # # Change the "pos" tweenable's "complex" tag to "nparray"
        # tags = self._state["pos"].tags
        # tags.remove("complex")
        # tags.add("nparray")
        # tags.add("fimage")

        # self.pos = pos

    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, value):
        self._pos = morpho.matrix.array(value)

    @property
    def orient(self):
        return self._orient

    @orient.setter
    def orient(self, value):
        self._orient = morpho.matrix.array(value)

    def copy(self):
        new = super().copy()
        new.orientable = self.orientable
        return new


    def primitives(self, camera): # orient=np.identity(3), focus=np.zeros(3)):
        if self.alpha == 0:
            return []

        orient = camera.orient
        focus = camera.focus

        if np.allclose(focus, 0):
            pos3d = orient @ self.pos
        else:
            pos3d = orient @ (self.pos - focus) + focus

        txt = Text()
        txt.text = self.text
        txt.pos = (pos3d[0] + 1j*pos3d[1]).tolist()
        txt.zdepth = pos3d[2]
        txt.size = self.size
        txt.transform = self.transform
        if self.orientable:
            txt.transform = (orient @ self.orient)[:2,:2] @ txt.transform
        txt.color = self.color
        txt.alpha = self.alpha
        txt.background = self.background
        txt.backAlpha = self.backAlpha
        txt.backPad = self.backPad
        txt.rotation = self.rotation
        txt.anchor_x = self.anchor_x
        txt.anchor_y = self.anchor_y

        txt.prescale_x = self.prescale_x
        txt.prescale_y = self.prescale_y

        txt.font = self.font
        txt.bold = self.bold
        txt.italic = self.italic

        return [txt]


    def draw(self, camera, ctx): #, orient=np.identity(3), focus=np.zeros(3)):
        primlist = self.primitives(camera)
        if len(primlist) == 0:
            return
        txt = primlist[0]
        txt.draw(camera, ctx)

Spacetext = SpaceText  # Synonym

# Multi version of the SpaceText class.
# See "SpaceText" and "MultiText" for more info.
class SpaceMultiText(MultiText):
    def __init__(self, text="", *args, **kwargs):
        if isinstance(text, str):
            textlist = [SpaceText(text, *args, **kwargs)]
        elif isinstance(text, list) or isinstance(text, tuple):
            textlist = [(SpaceText(item, *args, **kwargs) if isinstance(item, str) else item) for item in text]
        else:
            textlist = [SpaceText(text, *args, **kwargs)]

        # Create frame figure
        super().__init__(textlist)

    def primitives(self, camera):
        primlist = []
        for fig in self.figures:
            primlist.extend(fig.primitives(camera))

        return primlist

    def draw(self, camera, ctx):
        for fig in self.primitives(camera):
            fig.draw(camera, ctx)


SpaceMultitext = Spacemultitext = SpaceMultiText  # Synonyms


# Non-drawable figure that contains numerical data and tools for
# displaying it.
#
# TWEENABLES
# number = Internal number (real number). Default: 0.0
#
# OTHER ATTRIBUTES
# decimal = When displaying, to what decimal point should it round?
#           Default: 0 (round to nearest integer)
# leftDigits = Minimum number of digits to display left of the
#              decimal point. Prefixes zeros if necessary.
#              Default: 0 (no extra zeros are prefixed)
# rightDigits = Maximum number of digits to display right of the
#               decimal point. Appends zeros if necessary.
#               Default: 0 (no extra zeros are appended)
# truncate = Boolean indicating whether to truncate at the decimal value
#            instead of rounding. Default: False (round, don't truncate)
class Number(morpho.Figure):
    def __init__(self, number=0.0, decimal=0, leftDigits=0, rightDigits=0, truncate=False):
        morpho.Figure.__init__(self)

        number = morpho.Tweenable("number", number, tags=["scalar"])
        self.update([number])

        self.decimal = decimal
        self.leftDigits = leftDigits
        self.rightDigits = rightDigits
        self.truncate = truncate

    def copy(self):
        new = morpho.Figure.copy(self)

        new.decimal = self.decimal
        new.leftDigits = self.leftDigits
        new.rightDigits = self.rightDigits
        return new

    def __str__(self):
        if self.truncate:
            num = truncate(self.number, self.decimal)
        else:
            num = round(self.number, self.decimal)
        # Special case num is an integer AND we're not to display
        # any trailing zeros.
        if num == int(num) and self.rightDigits == 0:
            string = str(int(num))
            string = "0"*(self.leftDigits - len(string)) + string
            return string
        num = float(num)
        string = str(abs(num))
        numLeftDigits = string.find(".")
        numRightDigits = len(string) - numLeftDigits - 1
        string = "0"*(self.leftDigits - numLeftDigits) + string \
               + "0"*(self.rightDigits - numRightDigits)

        if num < 0: string = "-" + string

        return string

    ### TWEEN METHODS ###

    # @morpho.TweenMethod
    # def tweenLinear(self, other, t):
    #     Num = self.copy()
    #     Num.number = morpho.numTween(self.number, other.number, t)
    #     return Num

    @morpho.TweenMethod
    def tweenZoom(self, other, t):
        # This tween method requires both numbers to be positive.
        # If not, default to the Instant tween method.
        if not(self.number > 0 and other.number > 0):
            return self.copy()

        Num = self.copy()
        a = self.number
        b = other.number
        Num.number = a*(b/a)**t

        return Num





import morpholib as morpho
import morpholib.tools.color, morpholib.grid, morpholib.matrix
from morpholib.tools.basics import *
from morpholib.matrix import mat

# Polygon and SpacePolygon can be accessed from the shapes
# submodule as well as grid.
from morpholib.grid import Polygon, SpacePolygon, Spacepolygon

import cairo
cr = cairo

import math, cmath
import numpy as np

I2 = np.identity(2)


# Decorator modifies the tween methods of the Spline class to support
# tweening between splines with different node counts.
def handleSplineNodeInterp(tweenmethod):
    def wrapper(self, other, t, *args, **kwargs):
        len_self = self.length()
        len_other = other.length()

        # Use standard tween if node counts are the same
        if len_self == len_other:
            return tweenmethod(self, other, t, *args, **kwargs)

        # Otherwise, do some interpolation!

        # If either self or other have no nodes, give up, throw error
        if len_self == 0 or len_other == 0:
            raise ValueError("Can't interpolate between empty path and non-empty path!")

        # If self has more nodes than other, artifically insert
        # nodes into a copy of other before tweening
        if len_self > len_other:
            othercopy = other.copy()
            othercopy.insertNodes(len_self - len_other)
            # other.seq = insertNodesUniformlyTo(other.seq, len_self-len_other)
            return tweenmethod(self, othercopy, t, *args, **kwargs)
        # Else other has more nodes, so insert extra nodes to a
        # copy of self before tweening
        else:
            selfcopy = self.copy()
            selfcopy.insertNodes(len_other - len_self)
            # selfcopy.seq = insertNodesUniformlyTo(selfcopy.seq, len_other-len_self)
            # return super(Path, selfcopy).tweenLinear(other, t)
            return tweenmethod(selfcopy, other, t, *args, **kwargs)

    return wrapper

# Cubic Bezier Spline figure.
# Each node of the spline has three associated components:
# node position, inhandle position, outhandle position.
# where "inhandle" and "outhandle" refer to the two tangent
# control points associated with any node. "inhandle" means the handle
# that controls the curve's trajectory entering the node, and "outhandle"
# controls the curve's trajectory when exiting the node.
# See node(), inhandle(), and outhandle() for more info.
#
# inhandles and outhandles can also take on infinite values, in which
# case they are interpreted as being reflections of their counterpart
# handle. This allows two handles to be linked together, where one
# handle is implicitly controlled by the other. For example,
# inhandle = inf means its position is implicitly taken to be the
# reflection of the corresponding outhandle position about the node's
# position. If both inhandle AND outhandle are infinite, they are
# implicitly treated as if they coincide with the node position.
# NOTE: Values must be an infinite type! nans do not work and may
# break certain aspects of the Spline class!
#
# TWEENABLES
# data = Complex-valued matrix where rows index the nodes of the spline
#        and columns index the control point type:
#        (node, inhandle, outhandle)
#        This is not usually set directly. Instead use the methods
#        node(), inhandle(), outhandle(), etc. to define these.
#        Default: None (empty matrix)
# start = Initial draw point; a number between 0 and 1 where 0 is
#         the initial node and 1 is the final node.
# end = Final draw point; a number between 0 and 1 where 0 is
#       the initial node and 1 is the final node.
# color = Spline color (RGB vector-like). Default: (1,1,1) (white)
# alpha = Opacity. Default: 1 (opaque)
# width = Spline stroke thickness (in pixels). Default: 3
# alphaEdge = Path opacity independent of fill. Default: 1 (opaque)
# fill = Interior fill color (RGB vector-like). Default: [1,0,0] (red)
# alphaFill = Interior opacity. Default: 0 (invisible)
# origin = Translation value (complex number). Default: 0 (complex number).
# rotation = Path rotation about origin point (radians). Default: 0
# transform = Transformation matrix applied after all else. Default: np.eye(2)
#
# OTHER ATTRIBUTES
# dash = Dash pattern. Works exactly like how it does in cairo. It's a list
#        of ints which are traversed cyclically and will alternatingly indicate
#        number of pixels of visibility and invisibility.
# deadends = Set of ints specifying indices of seq that are "deadends". Meaning
#            no line segment will be drawn from the deadend index to the next index.
#            This is mainly used under the hood by helper functions like mathgrid()
#            to speed up rendering.
# showTangents = Boolean indicating whether to draw tangent line segments
#                at the node points of the spline. This is mainly for
#                debugging use while creating an animation.
#                Final animations should usually have showTangents = False.
#                By default, showTangents = False
class Spline(morpho.Figure):

    def __init__(self, data=None, width=3, color=(1,1,1), alpha=1):
        if data is None:
            # data = np.array([
            #     [0,-1j,oo],
            #     [1,1+1j,oo]
            #     ], dtype=complex)
            data = np.array([], dtype=complex)

        morpho.Figure.__init__(self)

        _data = morpho.Tweenable(name="_data", value=np.array(data, dtype=complex), tags=["nparray"])
        start = morpho.Tweenable(name="start", value=0, tags=["scalar"])
        end = morpho.Tweenable(name="end", value=1, tags=["scalar"])
        color = morpho.Tweenable(name="color", value=color, tags=["color"])
        alphaEdge = morpho.Tweenable(name="alphaEdge", value=1, tags=["scalar"])
        fill = morpho.Tweenable(name="fill", value=[1,0,0], tags=["color", "gradientfill", "nolinear", "nospiral"])
        alphaFill = morpho.Tweenable(name="alphaFill", value=0, tags=["scalar"])
        alpha = morpho.Tweenable(name="alpha", value=alpha, tags=["scalar"])
        width = morpho.Tweenable(name="width", value=width, tags=["size"])
        # headSize = morpho.Tweenable("headSize", 0, tags=["scalar"])
        # tailSize = morpho.Tweenable("tailSize", 0, tags=["scalar"])
        # outlineWidth = morpho.Tweenable("outlineWidth", value=0, tags=["size"])
        # outlineColor = morpho.Tweenable("outlineColor", value=[0,0,0], tags=["color"])
        # outlineAlpha = morpho.Tweenable("outlineAlpha", value=1, tags=["scalar"])
        origin = morpho.Tweenable("origin", value=0, tags=["complex", "nofimage"])
        rotation = morpho.Tweenable("rotation", value=0, tags=["scalar"])
        _transform = morpho.Tweenable("_transform", np.identity(2), tags=["nparray"])

        self.update([_data, start, end, color, alphaEdge, fill, alphaFill, alpha,
            width, origin, rotation, _transform]
            )


        # The dash pattern for this line. The format is identical to how
        # pycairo handles dash patterns: each item in the list is how long
        # ON and OFF dashes are, where the list is read cyclically.
        # Defaults to [] which means make the line solid.
        # Note that specifying only one value to the dash list is interpreted
        # as alternating that dash width ON and OFF.
        # Also note that dash pattern is ignored if gradient colors are used.
        self.dash = []

        # Set of indices that represent where a path should terminate.
        self.deadends = set()

        # Boolean indicates whether the control point tangents
        # should be shown. This is mainly for debugging purposes.
        self.showTangents = False


    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = np.array(value, dtype=complex)


    @property
    def transform(self):
        return self._transform

    @transform.setter
    def transform(self, value):
        self._transform = morpho.matrix.array(value)


    def copy(self):
        new = super().copy()
        new.dash = self.dash.copy() if "copy" in dir(self.dash) else self.dash
        new.deadends = self.deadends.copy()
        new.showTangents = self.showTangents

        return new

    # Returns the node count of the spline
    def length(self):
        return self.data.shape[0]

    # Returns or sets the position of the node of given index.
    # Usage: myspline.node(n) -> position of nth node
    #        myspline.node(n, value) sets nth node position to value
    def node(self, index, value=None):
        if value is None:
            return self.data[index, 0].tolist()
        self.data[index, 0] = value

    # Returns or sets the position of the inward handle
    # of the node at the given index (see node() for more info).
    # Input and output values of this method are in absolute
    # physical coordinates of the plane (as a complex number).
    # See also: inhandleRelative().
    #
    # If optional "raw" is set to True, the values are ripped
    # straight from the data array, including inf values;
    # but if set to False, it computes the current position
    # inf corresponds to.
    def inhandle(self, index, value=None, raw=False):
        if value is None:
            if raw:
                return self.data[index, 1].tolist()
            else:
                p, pin, pout = self.data[index,:].tolist()
                pin, pout = replaceInfHandles(p, pin, pout)
                return pin
        # Convert to oo given any non-finite value
        if isbadnum(value):
            value = oo
        self.data[index, 1] = value

    # Returns or sets the position of the outward handle
    # of the node at the given index (see node() for more info).
    # Input and output values of this method are in absolute
    # physical coordinates of the plane (as a complex number).
    # See also: outhandleRelative().
    #
    # If optional "raw" is set to True, the values are ripped
    # straight from the data array, including inf values;
    # but if set to False, it computes the current position
    # inf corresponds to.
    def outhandle(self, index, value=None, raw=False):
        if value is None:
            if raw:
                return self.data[index, 2].tolist()
            else:
                p, pin, pout = self.data[index,:].tolist()
                pin, pout = replaceInfHandles(p, pin, pout)
                return pout
        # Convert to oo given any non-finite value
        if isbadnum(value):
            value = oo
        self.data[index, 2] = value

    # Returns (node, inhandle, and outhandle) of given index.
    # Equivalent to extracting a row of the data array, but
    # converts any inf handle values into their current corresponding
    # positions like how inhandle() and outhandle() would output
    # by default.
    # Optionally set argument raw=True to make it actually return
    # just a raw copy of a given row of the data array
    # (converted into python list).
    def nodeData(self, index, raw=False):
        # p = self.node(index)
        # pin = self.inhandle(index)
        # pout = self.outhandle(index)

        p, pin, pout = self._data[index,:].tolist()
        if not raw:
            pin, pout = replaceInfHandles(p, pin, pout)

        return [p, pin, pout]

    # Returns or sets the position of the inward handle
    # of the node at the given index relative to the node position.
    # See also: inhandle()
    # Equivalent names for this method:
    # inhandlerelative, inhandleRel, inhandlerel
    #
    # If optional "raw" is set to True, the values are ripped
    # straight from the data array, including inf values;
    # but if set to False, it computes the current position
    # inf corresponds to.
    def inhandleRelative(self, index, value=None, raw=False):
        if value is None:
            return self.inhandle(index, value, raw) - self.node(index)
        # Convert to oo given any non-finite value
        if isbadnum(value):
            value = oo
        self.data[index, 1] = self.data[index, 0] + value

    inhandlerel = inhandleRel = inhandlerelative = inhandleRelative

    # Returns or sets the position of the outward handle
    # of the node at the given index relative to the node position.
    # See also: outhandle()
    # Equivalent names for this method:
    # outhandlerelative, outhandleRel, outhandlerel
    #
    # If optional "raw" is set to True, the values are ripped
    # straight from the data array, including inf values;
    # but if set to False, it computes the current position
    # inf corresponds to.
    def outhandleRelative(self, index, value=None, raw=False):
        if value is None:
            return self.outhandle(index, value, raw) - self.node(index)
            # return (self.data[index,2] - self.data[index,0]).tolist()
        # Convert to oo given any non-finite value
        if isbadnum(value):
            value = oo
        self.data[index, 2] = self.data[index, 0] + value

    outhandlerel = outhandleRel = outhandlerelative = outhandleRelative


    # Creates a new node at the specified point.
    # Optionally also specify inhandle and outhandle which default to inf.
    # Also optionally specify where to insert the node in the sequence.
    # By default, places it after the current final node.
    def newNode(self, point, inhandle=oo, outhandle=oo, beforeIndex=oo):
        beforeIndex = min(beforeIndex, self.length())
        if beforeIndex < 0:
            beforeIndex = beforeIndex % self.length()

        if self.length() == 0:
            self._data = np.array([[point, inhandle, outhandle]], dtype=complex)
        else:
            # self._data = np.vstack((self.data, [point,inhandle,outhandle]))
            self._data = np.insert(self.data, beforeIndex, [point,inhandle,outhandle], axis=0)

    # Deletes the node at the specified index. The dangling nodes on
    # either side will then be connected assuming they aren't prevented
    # by deadends.
    # NOTE: Calling delNode() will NOT update the deadends attribute!
    def delNode(self, index):
        self._data = np.delete(self._data, index, axis=0)


    # Closes the path IN PLACE if it is not already closed.
    def close(self):
        if self.length() < 2 or self.node(0) == self.node(1):
            return
        self._data = np.insert(self._data, self.length(), self._data[0].copy(), axis=0)


    # Returns the interpolated position along the path corresponding to the
    # parameter t, where t = 0 is the path start and t = 1 is the path end.
    # NOTE: This method ignores deadends and the transformation tweenables
    # origin, rotation, transform.
    def positionAt(self, t):
        length = self.length()
        if length < 2:
            raise IndexError("Spline must have at least 2 nodes!")
        segCount = length - 1
        T = t*segCount  # Global parameter value

        # Round to nearest node if within a billionth of it.
        tol = 1e-9
        if abs(T - round(T)) < tol:
            T = round(T)
            return self.node(T)

        index = int(T)  # Latest preceding node index
        param = T - index  # Local parameter value

        p0 = self.node(index)
        p1 = self.outhandle(index)
        p2 = self.inhandle(index+1)
        p3 = self.node(index+1)
        return morpho.bezierInterp(p0, p1, p2, p3, param)

    # Splits the spline at the parameter t (ranging from 0 to 1)
    # by inserting a new node at whatever point t corresponds to,
    # and auto-adjusting the handles of the adjacent nodes so that
    # the spline's shape is unchanged.
    # If optional "force" set to True, this method will split
    # the spline even at a node point. By default: force=False.
    def splitAt(self, t, force=False):
        segCount = self.length() - 1
        I = t*segCount  # Index value (possibly non-int)
        self.splitAtIndex(I, force=force)

        # length = self.length()
        # if length < 2:
        #     raise IndexError("Spline must have at least 2 nodes!")
        # segCount = length - 1
        # T = t*segCount  # Global parameter value

        # # Round to nearest node if within a billionth of it.
        # tol = 1e-9
        # if abs(T - round(T)) < tol:
        #     if force:
        #         T = round(T)
        #     else:
        #         # Do nothing if asked to split at a node point
        #         # and we're not forcing a split.
        #         return

        # index = int(T)  # Latest preceding node index
        # param = T - index  # Local parameter value
        # p0 = self.node(index)
        # p1 = self.outhandle(index)
        # p2 = self.inhandle(index+1)
        # p3 = self.node(index+1)
        # slice1, slice2 = morpho.bezier.splitBezier(p0, p1, p2, p3, param)

        # # Commit the handles of index and index+1 to ensure seamlessness
        # self.commitHandles(index, index+1)

        # # Modify starting and ending handles
        # self.outhandle(index, slice1[1])
        # self.inhandle(index+1, slice2[2])

        # # Define new node
        # p = slice2[0]
        # pin = slice1[2]
        # pout = slice2[1]

        # # Insert new node into data array between index and index+1
        # self._data = np.insert(self.data, index+1, [p,pin,pout], axis=0)

    # Similar to splitAt() except the input is a (possibly) non-integer
    # index value instead of a normalized parameter t between 0 and 1.
    # The given index can be any real number between 0 and
    # myspline.length()-1
    def splitAtIndex(self, I, force=False):
        length = self.length()
        if length < 2:
            raise IndexError("Spline must have at least 2 nodes!")
        segCount = length - 1

        # Round to nearest node if within a billionth of it.
        tol = 1e-9
        if abs(I - round(I)) < tol:
            if force:
                I = round(I)
                if I == segCount:
                    # self._data = np.insert(self._data, length, self._data[-1,:].copy(), axis=0)
                    self.outhandleRelative(-1, 0)
                    self.newNode(self.node(-1))
                    return
            else:
                # Do nothing if asked to split at a node point
                # and we're not forcing a split.
                return

        index = int(I)  # Latest preceding node index
        param = I - index  # Local parameter value
        p0 = self.node(index)
        p1 = self.outhandle(index)
        p2 = self.inhandle(index+1)
        p3 = self.node(index+1)
        slice1, slice2 = morpho.bezier.splitBezier(p0, p1, p2, p3, param)

        # Commit the handles of index and index+1 to ensure seamlessness
        self.commitHandles(index, index+1)

        # Modify starting and ending handles
        self.outhandle(index, slice1[1])
        self.inhandle(index+1, slice2[2])

        # Define new node
        p = slice2[0]
        pin = slice1[2]
        pout = slice2[1]

        # Insert new node into data array between index and index+1
        self._data = np.insert(self._data, index+1, [p,pin,pout], axis=0)


    # NOT IMPLEMENTED!
    # Extract a subspline.
    # a and b are indices, but can also be non-integers.
    # Actually, not sure if I want to implement this.
    # Gonna keep it low priority for now.
    def sub(self, a, b):
        raise NotImplementedError


    # Inserts the specified number of additional nodes to the
    # spline by interpolating along it. For large numNodes, the new
    # nodes will tend to bunch around the original nodes.
    def insertNodes(self, numNodes):
        segCount = self.length() - 1
        for n in range(numNodes):
            x = (n+1)/(numNodes+1) * segCount
            t = (x + n)/(segCount + n)
            self.splitAt(t, force=True)

    def _insertNodesUniformly_old(self, numNodes):
        raise NotImplementedError
        # Length and segment count of original spline
        length = self.length()
        segCount = length - 1
        tol = 1e-9  # Snap to nearest int if within this

        # Original version of the data array which will not
        # undergo modification during the following loop
        data = self.data.copy()

        for n in range(numNodes):
            # Parameter value in original index space
            t = (n+1)/(numNodes+1) * segCount

            # Round to nearest node if within tolerance.
            # Helps prevent floating point precision problems.
            if abs(t - round(t)) < tol:
                t = round(t)

            # Latest node index from ORIGINAL data array coming before t
            index = int(t)
            # Represents the actual index that "index" corresponds to in
            # the CURRENT self.data array. This is because we are modifying
            # self.data in place by inserting a new node each loop, so all
            # future indices get shifted up by 1 each cycle.
            INDEX = index + n

            p, pin, pout = data[index,:].tolist()
            pin, pout = replaceInfHandles(p, pin, pout)

            q, qin, qout = data[index+1,:].tolist()
            qin, qout = replaceInfHandles(q, qin, qout)

            # p, pin, pout = self.nodeData(INDEX)
            # q, qin, qout = self.nodeData(INDEX+1)

            slice1, slice2 = morpho.bezier.splitBezier(p, pout, qin, q, t-index)

            # Commit the handles of index and index+1 to ensure seamlessness
            self.commitHandles(INDEX, INDEX+1)

            # Modify starting and ending handles
            self.outhandle(INDEX, slice1[1])
            self.inhandle(INDEX+1, slice2[2])

            # Insert new node
            self.newNode(point=slice2[0], inhandle=slice1[2], outhandle=slice2[1], beforeIndex=INDEX+1)



    # # NOT TESTED YET!
    # # Returns the reflection of the specified inhandle about its
    # # corresponding node. If inhandle is inf, returns outhandle.
    # # If inhandle AND outhandle are inf, returns inf.
    # def inhandleReflection(self, index):
    #     inhandle = self.inhandle(index)
    #     if isbadnum(inhandle):
    #         outhandle = self.outhandle(index)
    #         return outhandle
    #     return reflect(inhandle, about=self.node(index))

    # # NOT TESTED YET!
    # # Returns the reflection of the specified outhandle about its
    # # corresponding node. If outhandle is inf, returns inhandle.
    # # If inhandle AND outhandle are inf, returns inf.
    # def outhandleReflection(self, index):
    #     outhandle = self.outhandle(index)
    #     if isbadnum(outhandle):
    #         inhandle = self.inhandle(index)
    #         return inhandle
    #     return reflect(outhandle, about=self.node(index))


    # Replaces any inf handles with the current positions they
    # would correspond to.
    # Optionally specify index or index range to commit
    # These index values can be negative, whereby they will be treated
    # cyclically like in python list indexing.
    # Example usage:
    # myspline.commitHandles()  # Commits all handles
    # myspline.commitHandles(n)  # Commits the handles of the nth node only
    # myspline.commitHandles(a,b)  # Commits the handles of nodes a thru b (inclusive)
    def commitHandles(self, index=None, upper=None):
        commitSplineHandles(self.data, index, upper)
        # if index is None and upper is None:
        #     index = 0
        #     upper = -1
        # elif index is None:
        #     index = 0
        # elif upper is None:
        #     upper = index

        # if index < 0:
        #     index = index % self.length()
        # if upper < 0:
        #     upper = upper % self.length()

        # for n in range(index, upper+1):
        #     pin, pout = replaceInfHandles(*self.data[n,:])
        #     self._data[n,1:] = pin, pout

    # Applies all of the transformation attributes
    # origin, rotation, transform
    # to the actual data array itself and then
    # resets the transformation attributes.
    def commitTransforms(self):
        # Rotate all points
        self._data *= cmath.exp(self.rotation*1j)

        # Perform linear transformation
        mat = morpho.matrix.Mat(self._transform)
        for i in range(self.length()):
            for j in range(3):
                z = self._data[i,j]
                if not isbadnum(z):
                    self._data[i,j] = mat*z

        # Translate
        self._data += self.origin

        # Convert any possible nans that were produced into infs.
        nan2inf(self._data)

        # Reset transformation tweenables to identities
        self.origin = 0
        self.rotation = 0
        self._transform = np.eye(2)


    def draw(self, camera, ctx):
        # Need at least two nodes to draw
        if self.data.shape[0] < 2:
            return

        # Check bounds of start and end
        if not(0 <= self.start <= 1):
            raise ValueError("start parameter must be in the range [0,1] (inclusive).")
        if not(0 <= self.end <= 1):
            raise ValueError("end parameter must be in the range [0,1] (inclusive).")

        # Handle trivial length path and start >= end.
        len_seq = self.length()
        maxIndex = len_seq - 1
        if maxIndex < 1 or self.start >= self.end or self.alpha == 0:
            return

        # If determinant of the transform matrix is too small,
        # don't attempt to draw.
        if abs(np.linalg.det(self._transform)) < 1e-6:
            return

        # Compute index bounds
        tol = 1e-9  # Snap to nearest integer if within this much of an integer.
        start = self.start*maxIndex
        if abs(start - round(start)) < tol:
            start = round(start)
        int_start = int(start)
        end = self.end*maxIndex
        if abs(end - round(end)) < tol:
            end = round(end)
        int_end = int(end)

        # Adjust based on deadends

        # Propagate start and int_start forward
        # if start is a deadend or in the void.
        while int_start in self.deadends:
            int_start += 1
            start = int_start

        # If end is STRICTLY in the void, floor it to int_end
        if end != int_end and int_end in self.deadends:
            end = int_end
        if end == int_end:
            # Backpropagate int_end until it is no longer a deadstart.
            while int_end-1 in self.deadends:
                int_end -= 1
            end = int_end

        # If start and end are too close, just skip
        if abs(start - end) < tol:
            return

        # If, after the adjustment, we get an empty path,
        # do nothing.
        if start >= end:
            return

        # If we have non-integer starting and ending indices,
        # we will need to split the spline at at least one point.
        needSplits = (start != int_start or end != int_end)

        if needSplits:
            # Save a copy of the original data array because
            # we will temporarily replace it.
            # This technique should probably be re-implemented better
            # in the future because it's expensive to copy an entire
            # np.array every single frame draw, though I expect it's
            # not too bad for a spline without a huge number of nodes.
            oldData = self._data
            self._data = oldData.copy()

            # Non-integer starting index
            if start != int_start:
                self.splitAtIndex(start)

                # Shift all relevant indices up one
                # to compensate for new node inserted early
                start += 1
                int_start += 1

                end += 1
                int_end += 1

            # Non-integer ending index
            if end != int_end:
                if int_start == int_end and start != int_start:
                    # Adjust end index value because it is now
                    # with respect to the newly added node instead
                    # of the original node.
                    end = int_end + (end-start)/(int_end+1-start)
                self.splitAtIndex(end)

        # Compute true initial and true final indices
        init = math.floor(start)
        final = math.ceil(end)

        # # Temporarily modify self.data in place to account
        # # for non-integer start and end
        # oldStart = self._data[int_start,:]
        # if start != init:
        #     p, pin, pout = self.nodeData(int_start)
        #     q, qin, qout = self.nodeData(int_start+1)
        #     m0, m1, m2, m3 = morpho.bezier.bezierLastSlice(p, pout, qin, q, start-int_start)

        #     self._data[int_start,:] = [m0, m1, 0]
        # if end != int_end:
        #     oldEnd = self._data[int_end+1,:]

        #     p, pin, pout = self._data[int_end,:] if int_end != int_start else oldStart
        #     q, qin, qout = self.nodeData(int_end+1)
        #     m0, m1, m2, m3 = morpho.bezier.bezierFirstSlice(p, pout, qin, q, end-int_end)

        #     self._data[int_end+1] = [m3, m2, 0]

        #     # self.seq[int_end+1] = morpho.numTween(
        #     #     self.seq[int_end] if int_end != int_start else oldStart, oldEnd, end-int_end
        #     #     )

        # Temporarily modify cairo coordinates to coincide with
        # physical coordinates.
        morpho.pushPhysicalCoords(camera.view, ctx)  # Contains a ctx.save()

        # Handle possible other transformations
        morpho.applyTransforms(ctx, self.origin, self.rotation, self.transform)


        # if start == init:
        # Initialize starting point
        zprev, inprev, outprev = self.data[init,:].tolist()
        inprev, outprev = replaceInfHandles(zprev, inprev, outprev)
        # else:
        #     p, pin, pout = self.nodeData(init)
        #     q, qin, qout = self.nodeData(init+1)
        #     m0, m1, m2, m3 = morpho.bezier.bezierLastSlice(p, pout, qin, q, start-init)
        #     zprev = m0
        #     outprev = m1

        # Move to starting point
        x,y = zprev.real, zprev.imag
        ctx.move_to(x,y)

        # Draw each curve
        # for n in range(self.data.shape[0]-1):
        for n in range(init, final):
            # Get next node, inhandle, and outhandle
            z, inhandle, outhandle = self.data[n+1,:].tolist()
            # Update handles based on possible inf values
            inhandle, outhandle = replaceInfHandles(z, inhandle, outhandle)

            x,y = z.real, z.imag

            # If previous node is a deadend, or current or previous
            # nodes are bad, move to next node.
            # Else, draw a curve to the next node.
            if n in self.deadends or isbadnum(z) or isbadnum(zprev):
                ctx.move_to(x,y)
            else:
                x1,y1 = outprev.real, outprev.imag
                x2,y2 = inhandle.real, inhandle.imag
                ctx.curve_to(x1,y1, x2,y2, x,y)

            # Update previous values to current values
            zprev = z
            # inprev = inhandle
            outprev = outhandle

        # # Handle non-integer ending index
        # if end != final:
        #     x1,y1 = outprev.real, outprev.imag

        #     p, pin, pout = self.nodeData(final)
        #     q, qin, qout = self.nodeData(final-1)
        #     m0, m1, m2, m3 = morpho.bezier.bezierFirstSlice(p, pout, qin, q, end-(final-1))
        #     x2,y2 = m2.real, m2.imag
        #     x,y = m3.real, m3.imag

        #     ctx.curve_to(x1,y1, x2,y2, x,y)

        # Handle gradients
        if self.alphaFill > 0:
            if isinstance(self.fill, morpho.color.GradientFill):
                self.fill.draw(camera, ctx, self.alphaFill*self.alpha, pushPhysicalCoords=False)
            # Handle normal colors
            else:
                ctx.set_source_rgba(*self.fill, self.alphaFill*self.alpha)
                ctx.fill_preserve()

        ctx.restore()

        # Set line width & color & alpha
        ctx.set_line_width(self.width)
        ctx.set_source_rgba(*self.color, self.alpha*self.alphaEdge)
        ctx.set_dash(self.dash)
        ctx.stroke()
        ctx.set_dash([])

        # Restore original data array if splits occurred
        if needSplits:
            self._data = oldData

        if self.showTangents:
            self.drawTangents(camera, ctx)

    def drawTangents(self, camera, ctx):
        # Need at least two nodes to draw
        if self.data.shape[0] < 2:
            return

        # If determinant of the transform matrix is too small,
        # don't attempt to draw.
        if abs(np.linalg.det(self.transform)) < 1e-6:
            return

        # Draw each tangent
        width = max(self.width/2, 1)
        for n in range(self.data.shape[0]):
            # Get next node, inhandle, and outhandle
            z, inhandle, outhandle = self.data[n,:].tolist()

            if isbadnum(z):
                continue

            # Update handles based on possible inf values
            inhandle, outhandle = replaceInfHandles(z, inhandle, outhandle)

            x,y = z.real, z.imag
            inx, iny = inhandle.real, inhandle.imag
            outx, outy = outhandle.real, outhandle.imag

            ctx.set_line_width(width)
            ctx.set_source_rgba(*self.color, self.alpha)

            # Temporarily modify cairo coordinates to coincide with
            # physical coordinates.
            morpho.pushPhysicalCoords(camera.view, ctx)  # Contains a ctx.save()

            # Handle possible other transformations
            morpho.applyTransforms(ctx, self.origin, self.rotation, self.transform)

            ctx.move_to(inx, iny)
            ctx.line_to(x,y)
            ctx.line_to(outx, outy)

            ctx.restore()
            ctx.stroke()

    # def draw(self, camera, ctx):
    #     self.draw0(camera, ctx, False)
    #     if self.showTangents:
    #         self.draw0(camera, ctx, True)


    # Converts the Spline figure to a similar-looking Path figure.
    # Optionally specify "segsteps" which is how many path steps to
    # use in a single bezier curve segment of the spline.
    # Defaults to 30 steps per segment.
    def toPath(self, segsteps=30):
        path = morpho.grid.line(0,1, steps=segsteps*(self.length()-1))

        # Make path follow the spline
        path = path.fimage(self.positionAt)

        # Match other tweenables
        path.start = self.start
        path.end = self.end
        path.color = self.color[:]
        path.alphaEdge = self.alphaEdge
        path.fill = self.fill.copy() if "copy" in dir(self.fill) else self.fill
        path.alphaFill = self.alphaFill
        path.alpha = self.alpha
        path.width = self.width
        path.origin = self.origin
        path.rotation = self.rotation
        path._transform = self._transform.copy()

        return path



    # NOT IMPLEMENTED YET!!!
    # Takes the function image of the spline under the given func.
    # Note: This applies to ALL control points: both nodes AND
    # handles. This function will skip any control points that are
    # inf or nan.
    #
    # Currently thinking I'll change the default behavior on
    # handle points so that the tangent lines change based on
    # the numerical DERIVATIVE of the given func. That seems
    # like it would produce the most faithful fimage of a spline.
    # However, that behavior perhaps could be suppressed by
    # changing a flag parameter.
    # It would basically work by computing the local linear
    # transformation about a node and applying that linear
    # transformation to both inhandle and outhandle locally.
    def fimage(self, func):
        raise NotImplementedError
        newfig = self.copy()
        # newfig._data = newfig._data.copy()

        for i in range(newfig.length()):
            for j in range(3):
                value = newfig.data[i,j].tolist()
                if not isbadnum(value):
                    newfig._data[i,j] = func(value)

        return newfig

    # Interpolates the data array of self with other, but after
    # replacing inf values with concrete values where necessary
    # (but leaves the data arrays of self and other unmodified).
    # This is done whenever one node has an inf handle, but its
    # counterpart handle in the other spline is finite.
    def splineDataInterp(self, other, t):
        data1 = self.data.copy()
        data2 = other.data.copy()

        crossCommitHandles(data1, data2)

        # Interpolate normally.
        data = morpho.lerp0(data1, data2, t)
        nan2inf(data)

        return data


    ### TWEEN METHODS ###

    @morpho.tweenMethod
    @morpho.color.handleGradientFills(["fill"])
    @handleSplineNodeInterp
    def tweenLinear(self, other, t):
        # Handle interpolating everything but the data tweenable
        tw = morpho.Figure.tweenLinear(self, other, t, ignore="_data")

        # Handle interpolating data
        tw._data = self.splineDataInterp(other, t)

        return tw

    @morpho.tweenMethod
    @morpho.color.handleGradientFills(["fill"])
    @handleSplineNodeInterp
    def tweenSpiral(self, other, t):
        # Handle interpolating everything but the data tweenable
        tw = morpho.Figure.tweenLinear(self, other, t, ignore="_data")

        # Handle interpolating data
        data1 = self.data.copy()
        data2 = other.data.copy()

        crossCommitHandles(data1, data2)

        # Interpolate using spiral tween method
        data12 = morpho.spiralInterpArray(data1, data2, t)
        nan2inf(data12)

        tw._data = data12

        return tw


    @classmethod
    def tweenPivot(cls, angle=tau/2):

        @morpho.TweenMethod
        @morpho.color.handleGradientFills(["fill"])
        @handleSplineNodeInterp
        def pivot(self, other, t):
            # Handle interpolating everything but the data tweenable
            tw = morpho.Figure.tweenLinear(self, other, t, ignore="_data")

            # Handle interpolating data
            data1 = self.data.copy()
            data2 = other.data.copy()

            crossCommitHandles(data1, data2)

            # Interpolate using pivot tween method
            data12 = morpho.pivotInterpArray(data1, data2, t, angle=angle)
            nan2inf(data12)

            tw._data = data12

            return tw

        return pivot


class SpaceSpline(Spline):
    def __init__(self, data=None, width=3, color=(1,1,1), alpha=1):

        # Use normal Spline constructor to start
        super().__init__(data=None, width=width, color=color, alpha=alpha)

        # Initialize origin triple
        origin = morpho.matrix.array([0,0,0])


        if data is None:
            # Use default data array supplied by superclass's constructor
            data = np.array([[[]]], dtype=float)
        elif type(data) is Spline:
            spline = data  # Rename so the following lines make more sense

            # Copy over state and all other attributes except data
            for name in self._state:
                if name != "_data":
                    self._state[name] = spline._state[name].copy()

            # Copy other attributes
            self.dash = spline.dash[:]
            self.deadends = spline.deadends.copy()
            origin = morpho.matrix.array(spline.origin)

            # Make "data" actually hold the 2D spline's data array
            # instead of mapping to the whole spline object.
            data = spline._data
        elif not isinstance(data, np.ndarray):
            raise TypeError("Unrecognized type for data array.")

        # Convert data array into a 3D array of real floats if
        # it lacks 3 dimensions
        if len(data.shape) < 3:
            # data3d = np.expand_dims(data.copy(), axis=2)

            # Convert data into empty 3D array of real floats
            # with same shape as data (just with 3 new slots)
            print(data.shape)
            Nrows, Ncols = data.shape
            data3d = np.zeros((Nrows, Ncols, 3))

            # Convert complex numbers into 3D vectors with
            # z-coordinate = 0.
            for i in range(Nrows):
                for j in range(Ncols):
                    s = data[i,j]
                    data3d[i,j,:] = [s.real, s.imag, 0]

            # Redefine data as data3d
            data = data3d

        # Convert all nans to infs before assigning the tweenable.
        nan2inf(data)
        _data = morpho.Tweenable("_data", data, tags=["nparray"])
        # Replace old _data tweenable that was inherited from
        # the superclass's constructor.
        self._state["_data"] = _data

        # Re-implement "origin" as a property so it will auto-convert
        # into np.array.
        self._state.pop("origin")
        _origin = morpho.Tweenable("_origin", origin, tags=["nparray", "nofimage"])
        self.extendState([_origin])

        # These transformation tweenables from 2D Spline are currently
        # not supported for SpaceSplines
        self._state.pop("rotation")
        self._state.pop("_transform")


    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = np.array(value, dtype=float)

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = morpho.matrix.array(value)

    # "transform" tweenable is currently unsupported for SpaceSpline
    @property
    def transform(self):
        raise AttributeError

    @transform.setter
    def transform(self, value):
        raise AttributeError

    # Returns or sets the position of the node of given index.
    # Usage: myspline.node(n) -> position of nth node
    #        myspline.node(n, value) sets nth node position to value
    def node(self, index, value=None):
        if value is None:
            return self.data[index, 0, :].copy()
        self.data[index, 0, :] = morpho.array(value)

    # Returns or sets the position of the inward handle
    # of the node at the given index (see node() for more info).
    # Input and output values of this method are in absolute
    # physical coordinates of the plane (as a complex number).
    # See also: inhandleRelative().
    #
    # If optional "raw" is set to True, the values are ripped
    # straight from the data array, including inf values;
    # but if set to False, it computes the current position
    # inf corresponds to.
    def inhandle(self, index, value=None, raw=False):
        if value is None:
            if raw:
                return self.data[index, 1, :]
            else:
                p, pin, pout = list(self.data[index,:,:])
                pin, pout = replaceInfHandles(p, pin, pout)
                return pin
        # Convert to oo given any non-finite value
        if isbadarray(value):
            value = oo
        else:
            value = morpho.array(value)
        self.data[index, 1, :] = value

    # Returns or sets the position of the outward handle
    # of the node at the given index (see node() for more info).
    # Input and output values of this method are in absolute
    # physical coordinates of the plane (as a complex number).
    # See also: outhandleRelative().
    #
    # If optional "raw" is set to True, the values are ripped
    # straight from the data array, including inf values;
    # but if set to False, it computes the current position
    # inf corresponds to.
    def outhandle(self, index, value=None, raw=False):
        if value is None:
            if raw:
                return self.data[index, 2, :]
            else:
                p, pin, pout = list(self.data[index,:,:])
                pin, pout = replaceInfHandles(p, pin, pout)
                return pout
        # Convert to oo given any non-finite value
        if isbadarray(value):
            value = oo
        else:
            value = morpho.array(value)
        self.data[index, 2, :] = value

    # Returns the matrix fully describing the node of given index.
    # Equivalent to extracting the 2D matrix slice at a specified first
    # index value: self.data[index,:,:]
    # but converts any inf handle values into their current
    # corresponding positions like how inhandle() and outhandle()
    # would output by default.
    # Optionally set argument raw=True to make it actually return
    # just a raw copy of a given node index of the data array.
    def nodeData(self, index, raw=False):
        if raw:
            return self._data[index,:,:].copy()

        p, pin, pout = list(self._data[index,:,:])
        pin, pout = replaceInfHandles(p, pin, pout)

        return np.array([p, pin, pout], dtype=float)

    # Returns or sets the position of the inward handle
    # of the node at the given index relative to the node position.
    # See also: inhandle()
    # Equivalent names for this method:
    # inhandlerelative, inhandleRel, inhandlerel
    #
    # If optional "raw" is set to True, the values are ripped
    # straight from the data array, including inf values;
    # but if set to False, it computes the current position
    # inf corresponds to.
    def inhandleRelative(self, index, value=None, raw=False):
        if value is None:
            return self.inhandle(index, value, raw) - self.node(index)
        # Convert to oo given any non-finite value
        if isbadarray(value):
            value = oo
        else:
            value = morpho.array(value)
        self.data[index, 1, :] = self.data[index, 0, :] + value

    inhandlerel = inhandleRel = inhandlerelative = inhandleRelative

    # Returns or sets the position of the outward handle
    # of the node at the given index relative to the node position.
    # See also: outhandle()
    # Equivalent names for this method:
    # outhandlerelative, outhandleRel, outhandlerel
    #
    # If optional "raw" is set to True, the values are ripped
    # straight from the data array, including inf values;
    # but if set to False, it computes the current position
    # inf corresponds to.
    def outhandleRelative(self, index, value=None, raw=False):
        if value is None:
            return self.outhandle(index, value, raw) - self.node(index)
        # Convert to oo given any non-finite value
        if isbadarray(value):
            value = oo
        else:
            value = morpho.array(value)
        self.data[index, 2, :] = self.data[index, 0, :] + value

    outhandlerel = outhandleRel = outhandlerelative = outhandleRelative

    # Creates a new node at the specified point.
    # Optionally also specify inhandle and outhandle which default to inf.
    # Also optionally specify where to insert the node in the sequence.
    # By default, places it after the current final node.
    def newNode(self, point, inhandle=(oo,oo,oo), outhandle=(oo,oo,oo), beforeIndex=oo):
        beforeIndex = min(beforeIndex, self.length())
        if beforeIndex < 0:
            beforeIndex = beforeIndex % self.length()

        # Convert to np.arrays if needed
        point = morpho.array(point)
        inhandle = morpho.array(inhandle)
        outhandle = morpho.array(outhandle)

        if self._data.size == 0:
            self._data = np.array([[point, inhandle, outhandle]], dtype=float)
        else:
            self._data = np.insert(self.data, beforeIndex, [point,inhandle,outhandle], axis=0)


    # Closes the path IN PLACE if it is not already closed.
    def close(self):
        if self.length() < 2 or np.array_equal(self.node(0), self.node(1)):
            return
        self._data = np.insert(self._data, self.length(), self._data[0,:,:].copy(), axis=0)


    # Translates the spline by the value of the "origin" attribute
    # and then resets the "origin" attribute to (0,0,0).
    def commitTransforms(self):
        self._data += self.origin
        self.origin = np.array([0,0,0], dtype=float)


    # Converts the Spline figure to a similar-looking SpacePath figure.
    # Optionally specify "segsteps" which is how many path steps to
    # use in a single bezier curve segment of the spline.
    # Defaults to 30 steps per segment.
    def toSpacePath(self, segsteps=30):
        steps = segsteps*(self.length()-1)
        seq = []
        for n in range(0,steps+1):
            seq.append(self.positionAt(n/steps))

        path = morpho.grid.SpacePath(seq)

        # Match other tweenables
        path.start = self.start
        path.end = self.end
        path.color = self.color[:]
        path.alphaEdge = self.alphaEdge
        path.fill = self.fill.copy() if "copy" in dir(self.fill) else self.fill
        path.alphaFill = self.alphaFill
        path.alpha = self.alpha
        path.width = self.width
        path.origin = self.origin.copy()

        return path

    toPath = toSpacepath = toSpacePath

    def primitives(self, camera):
        # If the spline is fully transparent, don't bother
        # creating any primitives. Just return the empty list.
        if self.alpha == 0:
            return []

        orient = camera.orient
        focus = camera.focus

        # Apply orient matrix transformation to all the vectors
        # along the final axis (axis 2)
        array = self._data
        if not np.allclose(self.origin, 0):
            # DON'T SIMPLIFY THIS LINE TO += !!!
            # We do NOT want this operation performed in place here!
            # It could end up modifying the original self._data!
            array = array + self.origin
        if not np.allclose(focus, 0):
            array = array - focus  # Do NOT simplify this to -= !!!
            array = np.tensordot(array, orient[:2,:], axes=((2),(1)))
            array += focus[:2]  # In place operation here is fine cuz array was replaced via arithmetic above
        else:
            array = np.tensordot(array, orient[:2,:], axes=((2),(1)))
        nan2inf(array)

        array2d = array[:,:,0] + 1j*array[:,:,1]
        spline = Spline(data=array2d)
        spline.start = self.start
        spline.end = self.end
        spline.color = self.color
        spline.alphaEdge = self.alphaEdge
        spline.fill = self.fill
        spline.alphaFill = self.alphaFill
        spline.alpha = self.alpha
        spline.width = self.width

        spline.dash = self.dash
        spline.deadends = self.deadends
        spline.showTangents = self.showTangents


        # zdepth of the whole spline is given by the median node's visual zdepth.
        max_index = self.length()-1
        x = (max_index) // 2  # This is (the floor of) the median index
        if max_index % 2 == 0:  # Even max index => easy median
            spline.zdepth = float((orient[2,:] @ (self._data[x,0,:]-focus)) + focus[2])
        else:  # Odd max index => average the two nearest
            w1, w2 = self._data[x:x+2, 0, :]
            spline.zdepth = float((orient[2,:] @ ((w1+w2)/2 - focus)) + focus[2])

        return [spline]


    def draw(self, camera, ctx):
        primlist = self.primitives(camera)
        if len(primlist) == 0:
            return
        path = primlist[0]
        path.draw(camera, ctx)

    def drawTangents(self, camera, ctx):
        raise NotImplementedError


    ### TWEEN METHODS ###

    def tweenSpiral(self, other, t):
        raise NotImplementedError

    @classmethod
    def tweenPivot(cls, angle=tau/2):
        raise NotImplementedError




# Helper mainly for Spline class. Computes the reflection of "point"
# about the base point called "about".
def reflect(point, about):
    return 2*about - point

# Given the data triplet describing a node in a spline,
# replaces inf handle values
# with the current position values they correspond to based on
# their partner handle.
# The results are returned as a pair (new_inhandle, new_outhandle)
# This function does NOT operate in place: the inputs are unmodified.
def replaceInfHandles(point, inhandle, outhandle):
    # Get booleans indicating badness of handles
    inbad, outbad = isbadarray(inhandle), isbadarray(outhandle)
    # Reflect handles if necessary
    if inbad and outbad:
        inhandle = outhandle = 1*point  # 1* makes a copy if it's np.array
    elif outbad:
        outhandle = reflect(inhandle, about=point)
    elif inbad:
        inhandle = reflect(outhandle, about=point)
    else:
        # This makes copies of np.arrays if they are np.arrays,
        # but leaves complex numbers unchanged.
        inhandle = 1*inhandle
        outhandle = 1*outhandle

    return inhandle, outhandle


# Given spline data array,
# replaces any inf handles with the current positions they
# would correspond to.
# Optionally specify index or index range to commit.
# NOTE: This function modifies the data array IN PLACE!
def commitSplineHandles(data, index=None, upper=None):
    length = data.shape[0]

    if index is None and upper is None:
        index = 0
        upper = -1
    elif index is None:
        index = 0
    elif upper is None:
        upper = index

    if index < 0:
        index = index % length
    if upper < 0:
        upper = upper % length

    for n in range(index, upper+1):
        pin, pout = replaceInfHandles(*data[n,:])
        data[n,1:] = (pin, pout)

# Given two data arrays of the same shape,
# checks for entries where one array has inf but the other has finite,
# and then replaces the infinite entry with its corresponding
# reflected finite value FOR ALL entries that share a row with that one.
# This operation is done IN PLACE to both data arrays.
# This function is mainly used to help tween two splines in which
# one spline has an implicit handle (inf) and the other has an explicit
# handle (finite).
def crossCommitHandles(data1, data2):
    # Find all rows where one entry is +oo but the other is finite.
    # These are the rows where we will have to replace the infinities
    # with concrete values.
    # rows = set(np.where(np.isinf(abs(data1) - abs(data2)).any(axis=1))[0].tolist())
    rows = set(np.where(np.isinf(abs(data1) - abs(data2)).any(axis=tuple(range(1,len(data1.shape)))))[0].tolist())
    for r in rows:
        commitSplineHandles(data1, r)
        commitSplineHandles(data2, r)

# def crossCommitHandles3D(data1, data2):
#     # Find all node matrices where one entry is +oo but the other
#     # is finite. These are the indices where we have to replace the
#     # inf values with concrete values.
#     nodeIndices = set(np.where(np.isinf(abs(data1) - abs(data2)).any(axis=(1,2)))[0].tolist())
#     for n in nodeIndices:
#         commitSplineHandles(data1, n)
#         commitSplineHandles(data2, n)


# Converts all entries of the array that are nan into inf
# IN PLACE!
def nan2inf(array):
    array[np.isnan(array)] = oo



# NOT IMPLEMENTED YET!!!
# Parses an SVG string that describes a Bezier spline and then
# returns a data array
def SVGdata(string):
    raise NotImplementedError



# A pure ellipse object independent of the Polygon class.
# Note: Cannot be tweened into a polygon.
#
# TWEENABLES
# pos = Ellipse center (complex number). Default: 0
# xradius, yradius = Horizontal and vertical radii (physical units). Default: 1
# strokeWeight = Border thickness (in pixels). Default: 3
# color = Border color (RGB list). Default: [1,1,1] (white)
# fill = Interior fill color (RGB list). Default [1,0,0] (red)
# edgeAlpha = Border opacity. Default: 1 (opaque)
# fillAlpha = Interior opacity. Default: 1 (opaque)
# alpha = Overall opacity. Multiplies edgeAlpha and fillAlpha.
#         Default: 1 (opaque)
class Ellipse(morpho.Figure):

    def __init__(self, pos=0, xradius=1, yradius=1,
        strokeWeight=3, color=(1,1,1), fill=(1,0,0),
        edgeAlpha=1, fillAlpha=1, alpha=1):

        super().__init__()

        pos = morpho.Tweenable("pos", pos, tags=["complex", "position"])
        xradius = morpho.Tweenable("xradius", xradius, tags=["scalar"])
        yradius = morpho.Tweenable("yradius", yradius, tags=["scalar"])
        strokeWeight = morpho.Tweenable("strokeWeight", strokeWeight, tags=["scalar"])
        color = morpho.Tweenable("color", list(color), tags=["color"])
        fill = morpho.Tweenable("fill", list(fill), tags=["color"])
        edgeAlpha = morpho.Tweenable("edgeAlpha", edgeAlpha, tags=["scalar"])
        fillAlpha = morpho.Tweenable("fillAlpha", fillAlpha, tags=["scalar"])
        alpha = morpho.Tweenable("alpha", alpha, tags=["scalar"])

        self.update([pos, xradius, yradius, strokeWeight, color, fill, edgeAlpha, fillAlpha, alpha])

    # NOT IMPLEMENTED YET!!!
    def toPolygon(self, dTheta=5):
        raise NotImplementedError

    def draw(self, camera, ctx):
        view = camera.view

        X,Y = morpho.screenCoords(self.pos, view, ctx)

        ctx.save()
        ctx.translate(X,Y)
        WIDTH = morpho.pixelWidth(self.xradius, view, ctx)
        HEIGHT = morpho.pixelHeight(self.yradius, view, ctx)

        # Zero is not allowed. Constrain to 0.1
        WIDTH = max(WIDTH, 0.1)
        HEIGHT = max(HEIGHT, 0.1)

        ctx.scale(WIDTH, HEIGHT)

        ctx.move_to(1,0)
        ctx.arc(0,0, 1, 0, tau)
        ctx.restore()

        ctx.set_source_rgba(*self.fill, self.fillAlpha*self.alpha)
        ctx.fill_preserve()
        ctx.set_source_rgba(*self.color, self.edgeAlpha*self.alpha)
        ctx.set_line_width(self.strokeWeight)
        ctx.stroke()


# Creates an arc of an ellipse.
# Angles should be interpreted as if the ellipse were a circle.
# That is, angles refer to a circular arc BEFORE being stretched
# into an ellipse. They are in units of radians.
#
# TWEENABLES
# pos = Ellipse center (complex number). Default: 0
# xradius, yradius = Horizontal and vertical radii (physical units). Default: 1
# theta0, theta1 = Angles (in rad) defining the angular span.
#                  The arc is always drawn starting from theta0 and going
#                  toward theta1, covering all angles between theta0 and theta1.
#                  Default: 0,2pi
# strokeWeight = Border thickness (in pixels). Default: 3
# color = Border color (RGB list). Default: [1,1,1] (white)
# alpha = Opacity. Default: 1 (opaque)
class EllipticalArc(morpho.Figure):

    def __init__(self, pos=0, xradius=1, yradius=1, theta0=0, theta1=tau,
        strokeWeight=3, color=(1,1,1), alpha=1):

        super().__init__()

        pos = morpho.Tweenable("pos", pos, tags=["complex", "position"])
        xradius = morpho.Tweenable("xradius", xradius, tags=["scalar"])
        yradius = morpho.Tweenable("yradius", yradius, tags=["scalar"])
        theta0 = morpho.Tweenable("theta0", theta0, tags=["scalar"])
        theta1 = morpho.Tweenable("theta1", theta1, tags=["scalar"])
        strokeWeight = morpho.Tweenable("strokeWeight", strokeWeight, tags=["scalar"])
        color = morpho.Tweenable("color", list(color), tags=["color"])
        alpha = morpho.Tweenable("alpha", alpha, tags=["scalar"])

        self.update([pos, xradius, yradius, theta0, theta1, strokeWeight, color, alpha])

    # Converts the figure into an equivalent Path figure.
    # Optionally specify the angular steps (in degs). Default: 5
    def toPath(self, dTheta=5):

        theta0, theta1 = self.theta0, self.theta1
        # If angular span is greater than tau,
        # just draw a circle
        if abs(theta1 - theta0) >= tau:
            theta1 = theta0 + tau
        # Ensure theta0 <= theta1
        elif theta1 < theta0:
            theta0, theta1 = theta1, theta0

        # steps = int(math.ceil(360 / abs(dTheta)))
        dTheta *= tau/360  # convert dTheta to radians
        steps = math.ceil((theta1 - theta0) / abs(dTheta))

        # Make unit circle arc
        z0 = cmath.exp(theta0*1j)
        z1 = cmath.exp(theta1*1j)
        seq = [z0]
        for n in range(1, steps):
            seq.append(cmath.exp((theta0+n*dTheta)*1j))
        seq.append(z1)

        path = morpho.grid.Path(seq)
        path.width = self.strokeWeight
        path.color = self.color[:]
        path.alpha = self.alpha

        # Stretch it into an ellipse and move it
        path = path.fimage(lambda z: mat(self.xradius,0,0,self.yradius)*z)
        path = path.fimage(lambda z: z + self.pos)

        return path


    def draw(self, camera, ctx):
        view = camera.view

        X,Y = morpho.screenCoords(self.pos, view, ctx)

        ctx.save()
        ctx.translate(X,Y)
        WIDTH = max(morpho.pixelWidth(self.xradius, view, ctx), 0.1)
        HEIGHT = max(morpho.pixelHeight(self.yradius, view, ctx), 0.1)
        ctx.scale(WIDTH, HEIGHT)

        theta0, theta1 = self.theta0, self.theta1
        # If angular span is greater than tau,
        # just draw a circle
        if abs(theta1 - theta0) >= tau:
            theta0 = 0
            theta1 = tau
        elif theta1 < theta0:
            theta0, theta1 = theta1, theta0

        Z0 = cmath.exp(theta0*1j)
        ctx.move_to(Z0.real, Z0.imag)
        ctx.arc(0,0, 1, theta0, theta1)
        ctx.restore()

        ctx.set_source_rgba(*self.color, self.alpha)
        ctx.set_line_width(self.strokeWeight)
        ctx.stroke()


# Creates a segment of an ellipse: a rectangle in polar coordinate space.
# Parameters are essentially identical to EllipticalArc, but with new parameter
# "innerFactor" which is a number between 0 and 1 representing where the inner arc
# should appear relative to the outer arc. So setting innerFactor = 0.25 means
# the inner arc will appear 25 percent of the way from the origin to the outer arc.
class Pie(EllipticalArc):
    def __init__(self, pos=0, xradius=1, yradius=1, innerFactor=0, theta0=0, theta1=tau,
        strokeWeight=3, color=(1,1,1), alphaEdge=1, fill=(1,0,0), alphaFill=1, alpha=1):

        super().__init__(pos, xradius, yradius, theta0, theta1, strokeWeight,
            color, alpha)

        innerFactor = morpho.Tweenable("innerFactor", innerFactor, tags=["scalar"])
        alphaEdge = morpho.Tweenable("alphaEdge", alphaEdge, tags=["scalar"])
        fill = morpho.Tweenable("fill", fill, tags=["color"])
        alphaFill = morpho.Tweenable("alphaFill", alphaFill, tags=["scalar"])
        self.extendState([innerFactor, alphaEdge, fill, alphaFill])

    def draw(self, camera, ctx):
        view = camera.view

        X,Y = morpho.screenCoords(self.pos, view, ctx)

        ctx.save()
        ctx.translate(X,Y)
        WIDTH = max(morpho.pixelWidth(self.xradius, view, ctx), 0.1)
        HEIGHT = max(morpho.pixelHeight(self.yradius, view, ctx), 0.1)
        ctx.scale(WIDTH, HEIGHT)

        theta0, theta1 = self.theta0, self.theta1
        # If angular span is greater than tau,
        # just draw a circle
        if abs(theta1 - theta0) >= tau:
            theta1 = theta0 + tau
        elif theta1 < theta0:
            theta0, theta1 = theta1, theta0

        Z0 = cmath.exp(theta0*1j)
        Z1 = cmath.exp(theta1*1j)
        W0 = self.innerFactor*Z0
        W1 = self.innerFactor*Z1
        ctx.move_to(Z0.real, Z0.imag)
        ctx.arc(0,0, 1, theta0, theta1)
        ctx.line_to(W1.real, W1.imag)
        ctx.arc_negative(0,0, self.innerFactor, theta1, theta0)
        ctx.close_path()
        ctx.restore()

        # Draw the fill
        ctx.set_source_rgba(*self.fill, self.alphaFill*self.alpha)
        ctx.fill_preserve()

        ctx.set_source_rgba(*self.color, self.alphaEdge*self.alpha)
        ctx.set_line_width(self.strokeWeight)
        ctx.stroke()

    def toPath(self, dTheta=5):
        raise NotImplementedError

    # Converts the figure into an equivalent polygon figure
    # Optionally specify the angular steps (in degs). Default: 5
    def toPolygon(self, dTheta=5):

        theta0, theta1 = self.theta0, self.theta1
        # If angular span is greater than tau,
        # just draw a circle
        if abs(theta1 - theta0) >= tau:
            theta1 = theta0 + tau
        # Ensure theta0 <= theta1
        elif theta1 < theta0:
            theta0, theta1 = theta1, theta0

        # steps = int(math.ceil(360 / abs(dTheta)))
        dTheta *= tau/360  # convert dTheta to radians
        steps = math.ceil((theta1 - theta0) / abs(dTheta))

        # Make unit circle
        z0 = cmath.exp(theta0*1j)
        z1 = cmath.exp(theta1*1j)
        w0 = self.innerFactor * z0
        w1 = self.innerFactor * z1
        seq = [z0]
        for n in range(1, steps):
            seq.append(cmath.exp((theta0+n*dTheta)*1j))
        seq.append(z1)

        if self.innerFactor > 0:
            innerSeq = seq[::-1]
            innerSeq = np.array(innerSeq, dtype=complex)
            innerSeq *= self.innerFactor
            innerSeq = innerSeq.tolist()
        else:
            innerSeq = [0]

        seq.extend(innerSeq)

        # Make the polygon
        poly = morpho.grid.Polygon(seq)

        # Style parameters
        poly.width = self.strokeWeight
        poly.color = self.color[:]
        poly.fill = self.fill[:]
        poly.alphaEdge = self.alphaEdge
        poly.alphaFill = self.alphaFill
        poly.alpha = self.alpha

        # Stretch it into an ellipse and move it
        poly = poly.fimage(lambda z: mat(self.xradius,0,0,self.yradius)*z)
        poly = poly.fimage(lambda z: z + self.pos)
        return poly



### GADGETS ###

# These functions can be found here as well as grid.
line = morpho.grid.line
spaceLine = morpho.grid.spaceLine
rect = morpho.grid.rect
arc = morpho.grid.arc




### SCRAPS ###

'''
The following is the old work-in-progress notes for developing the
Spline class. I'm keeping it for now because it had a lot of ideas
and I'm not completely sure yet if I want to just discard it all.

NOT FULLY IMPLEMENTED YET!
    Cubic Bezier spline figure.
    We have a tweenable called "data" which is a list of triplets, where
    each triplet (called a node) describes a single node along the spline.
    node[0] = Position of node
    node[1] = Inward control point/handle
    node[2] = Outward control point/handle
    For the first node triplet, the inward handle is ignored, and similarly
    for the last node triplet, the outward handle is ignored.
    The handles can also take on "nan" values, which means when drawn, they
    will be interpreted as reflections of the complement handle, or if its
    complement is nan, it will be set equal to the node position.
    QUESTION: Should "data" be a python matrix? or a np.array?
    Currently thinking np.array is best. See fimage() discussion below.
    - Also, it would make extending this class to SpaceSpline easier, I think.
      tweenLinear() will treat all np.arrays alike, and similarly for fimage()

    Also consider having a "deadends" attribute which indicates which splines
    end at a deadend.

    Double-check that tweenLinear() can properly handle what will
    essentially be a python matrix of complex numbers! Also carefully
    check over how it will handle the positions being np.arrays when
    we eventually implement this in 3D.

    Perhaps have some helper methods/properties that allow the user to
    edit the spline data more easily?
    - Like these: They return the corresponding value unless value is
      other than python None, in which case, it SETS the value and the
      function returns python None
      + node(nodenum, value=None)
      + inhandle(nodenum, value=None)
      + outhandle(nodenum, value=None)
    - Maybe also have a method that can parse an SVG description of a
      spline path? Maybe implement this as part of the constructor?
      But have the meat of the SVG parser in its own outside helper
      function.

    Also, how should fimage() work? Currently thinking default behavior is fine,
    meaning handles are subjected to fimage() just like nodes are. If this is
    the desired behavior, then I think implementing the "data" tweenable as a
    complex-valued np.array makes the most sense.
    - However, you might slightly modify fimage() so that it just skips nan
      values instead of evaluating the function on them.
'''



'''
A collection of figure classes that can be useful for animating
Calculus concepts.
'''

import morpholib as morpho
mo = morpho
import morpholib.anim, morpholib.graph, morpholib.grid, morpholib.matrix
from morpholib.tools.basics import *

import numpy as np
from cmath import exp

deg90 = tau/4
deg270 = 3*tau/4
rot90x = morpho.matrix.rotation([1,0,0], tau/4)
rot270y = morpho.matrix.rotation([0,1,0], -tau/4)
swapxy = morpho.array([[0,1,0],[1,0,0],[0,0,1]])


class IntegralArea(morpho.Figure):

    def __init__(self, func=lambda x: x, start=0, end=1,
        strokeWeight=3, color=(0,0,0), fill=(1,0,0), alpha=1,
        steps=50):

        # Construct default figure
        super().__init__()

        func = morpho.Tweenable("func", func, tags=["function"])
        start = morpho.Tweenable("start", start, tags=["scalar"])
        end = morpho.Tweenable("end", end, tags=["scalar"])
        color = morpho.Tweenable("color", list(color), tags=["color"])
        fill = morpho.Tweenable("fill", list(fill), tags=["color"])
        alpha = morpho.Tweenable("alpha", alpha, tags=["scalar"])
        strokeWeight = morpho.Tweenable("strokeWeight", strokeWeight, tags=["scalar"])
        steps = morpho.Tweenable("steps", steps, tags=["integer"])

        self.update([func, start, end, color, fill, alpha, strokeWeight, steps])

        # # Integrand function
        # self.func = func

        # Internal polygon
        self.polygon = morpho.grid.Polygon()
        self.polygon.alphaFill = 1
        self.updatePolygon()

    def updatePolygon(self):
        path = morpho.graph.realgraph(
            self.func, self.start, self.end, self.steps,
            self.strokeWeight, self.color[:]
            )

        self.polygon.vertices = path.seq + [path.seq[-1].real, path.seq[0].real]
        self.polygon.width = self.strokeWeight
        self.polygon.color = self.color[:]
        self.polygon.fill = self.fill[:]
        self.polygon.alpha = self.alpha

    def copy(self, *args, **kwargs):
        # Copy according to superclass first
        copy = super().copy(*args, **kwargs)

        # # Now copy the function and internal polygon
        # copy.func = self.func
        # Actually, not sure we need to copy the internal
        # polygon, because I think a new one will be made
        # when the generic copy() calls the constructor.
        copy.polygon = self.polygon.copy()

        return copy

    def draw(self, camera, ctx):
        self.updatePolygon()
        self.polygon.draw(camera, ctx)

class DoubleIntegralVolume(morpho.SpaceFigure):

    def __init__(
        self, func=lambda v: v, mode="dydx",
        inmin=0, inmax=1, outmin=0, outmax=1,
        width=0, color=(0,0,0), alphaEdge=1,
        fill=(1,0,0), alphaFill=1, alpha=1,
        steps=24, origin=morpho.array([0,0,0])
        ):

        # Construct default figure
        super().__init__()

        mode = mode.lower()
        if not(mode == "dxdy" or mode == "dydx"):
            raise ValueError("mode must be 'dydx' or 'dxdy'")

        # Modify constants into constant functions for the
        # inner bounds.
        if not isinstance(inmin, morpho.function):
            inmin = lambda t, value=inmin: value
        if not isinstance(inmax, morpho.function):
            inmax = lambda t, value=inmax: value

        # Define tweenables
        # innerBounds = morpho.Tweenable("innerBounds", innerBounds, tags=["function", "list", "loop"])
        # outerBounds = morpho.Tweenable("outerBounds", outerBounds, tags=["scalar", "list"])

        func = morpho.Tweenable("func", func, tags=["function"])
        _inmin = morpho.Tweenable("_inmin", inmin, tags=["function"])
        _inmax = morpho.Tweenable("_inmax", inmax, tags=["function"])
        outmin = morpho.Tweenable("outmin", outmin, tags=["scalar"])
        outmax = morpho.Tweenable("outmax", outmax, tags=["scalar"])
        width = morpho.Tweenable("width", width, tags=["scalar"])
        # strokeWeight = morpho.Tweenable("strokeWeight", strokeWeight, tags=["scalar"])
        color = morpho.Tweenable("color", color, tags=["color"])
        alphaEdge = morpho.Tweenable("alphaEdge", alphaEdge, tags=["scalar"])
        alphaFill = morpho.Tweenable("alphaFill", alphaFill, tags=["scalar"])
        fill = morpho.Tweenable("fill", fill, tags=["color"])
        alpha = morpho.Tweenable("alpha", alpha, tags=["scalar"])
        steps = morpho.Tweenable("steps", steps, tags=["integer"])
        _origin = morpho.Tweenable("_origin", morpho.array(origin), tags=["nparray"])

        self.update([func, _inmin, _inmax, outmin, outmax,
            width, color,
            alphaEdge, alphaFill, fill, alpha, steps, _origin])

        # Non-tweenable attributes
        self.mode = mode
        self.shading = True

    @property
    def inmin(self):
        return self._inmin

    @inmin.setter
    def inmin(self, value):
        if not isinstance(value, morpho.function):
            self._inmin = lambda t, value=value: value
        else:
            self._inmin = value

    @property
    def inmax(self):
        return self._inmax

    @inmax.setter
    def inmax(self, value):
        if not isinstance(value, morpho.function):
            self._inmax = lambda t, value=value: value
        else:
            self._inmax = value


    def copy(self):
        new = super().copy()
        new.mode = self.mode
        new.shading = self.shading
        return new

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = morpho.array(value)

    def primitives(self, camera):

        a,b = self.outmin, self.outmax
        xLen = b-a
        cfunc, dfunc = self.inmin, self.inmax
        dr = 1/self.steps
        swap = (self.mode.lower()=="dxdy")

        westWall = morpho.grid.quadgrid([0,1,0,1],
            dx=1, dy=dr,
            width=self.width, color=self.color, alphaEdge=self.alphaEdge,
            fill=self.fill, alphaFill=self.alphaFill, alpha=self.alpha
            )
        westWall.shading = self.shading
        westWall = westWall.fimage(lambda v: rot270y @ v)
        westc,westd = cfunc(a), dfunc(a)
        westYlen = westd-westc
        def stretchWest(v):
            x,y,z = v
            X = a
            Y = y*westYlen + westc
            Z = z

            return morpho.array([X,Y,Z])

        westWall = westWall.fimage(stretchWest)
        if swap:
            westWall = westWall.fimage(lambda v: swapxy @ v)
        def curveWest(v):
            x,y,z = v
            X,Y = x,y
            Z = z*self.func(v)[2]

            return morpho.array([X,Y,Z])

        westWall = westWall.fimage(curveWest)


        eastWall = morpho.grid.quadgrid([0,1,0,1],
            dx=1, dy=dr,
            width=self.width, color=self.color, alphaEdge=self.alphaEdge,
            fill=self.fill, alphaFill=self.alphaFill, alpha=self.alpha
            )
        eastWall.shading = self.shading
        eastWall = eastWall.fimage(lambda v: rot270y @ v)
        eastc,eastd = cfunc(b), dfunc(b)
        eastYlen = eastd-eastc
        def stretchEast(v):
            x,y,z = v
            X = b
            Y = y*eastYlen + eastc
            Z = z

            return morpho.array([X,Y,Z])

        eastWall = eastWall.fimage(stretchEast)
        if swap:
            eastWall = eastWall.fimage(lambda v: swapxy @ v)
        def curveEast(v):
            x,y,z = v
            X,Y = x,y
            Z = z*self.func(v)[2]

            return morpho.array([X,Y,Z])

        eastWall = eastWall.fimage(curveEast)


        southWall = morpho.grid.quadgrid([0,1,0,1],
            dx=dr, dy=1,
            width=self.width, color=self.color, alphaEdge=self.alphaEdge,
            fill=self.fill, alphaFill=self.alphaFill, alpha=self.alpha
            )
        southWall.shading = self.shading
        southWall = southWall.fimage(lambda v: rot90x @ v)
        def stretchSouth(v):
            x,y,z = v
            X = xLen*x + a
            Y = cfunc(X)
            Z = z

            return morpho.array([X,Y,Z])

        southWall = southWall.fimage(stretchSouth)
        if swap:
            southWall = southWall.fimage(lambda v: swapxy @ v)
        def southTop(v):
            x,y,z = v
            X,Y = x,y
            Z = z*self.func(v)[2]

            return morpho.array([X,Y,Z])
        southWall = southWall.fimage(southTop)


        northWall = morpho.grid.quadgrid([0,1,0,1],
            dx=dr, dy=1,
            width=self.width, color=self.color, alphaEdge=self.alphaEdge,
            fill=self.fill, alphaFill=self.alphaFill, alpha=self.alpha
            )
        northWall.shading = self.shading
        northWall = northWall.fimage(lambda v: rot90x @ v)
        def stretchNorth(v):
            x,y,z = v
            X = xLen*x + a
            Y = dfunc(X)
            Z = z

            return morpho.array([X,Y,Z])

        northWall = northWall.fimage(stretchNorth)
        if swap:
            northWall = northWall.fimage(lambda v: swapxy @ v)
        def northTop(v):
            x,y,z = v
            X,Y = x,y
            Z = z*self.func(v)[2]

            return morpho.array([X,Y,Z])
        northWall = northWall.fimage(northTop)

        # if self.strokeWeight > 0 and self.alphaEdge > 0:
        #     # Construct edges at the four corners
        #     SWedge = morpho.grid.SpacePath([[a,westc,0],[a,westc,self.func(mo.array([a,westc,0]))[2]] ])
        #     SEedge = morpho.grid.SpacePath([[b,eastc,0],[b,eastc,self.func(mo.array([b,eastc,0]))[2]] ])
        #     NEedge = morpho.grid.SpacePath([[b,eastd,0],[b,eastd,self.func(mo.array([b,eastd,0]))[2]] ])
        #     NWedge = morpho.grid.SpacePath([[a,westd,0],[a,westd,self.func(mo.array([a,westd,0]))[2]] ])

        #     # Assign common style parameters
        #     edges = [SWedge, SEedge, NEedge, NWedge]
        #     for fig in edges:
        #         fig.color = self.color
        #         fig.width = self.strokeWeight
        #         fig.alpha = self.alphaEdge*self.alpha
        # else:
        #     edges = []

        prims = []
        for fig in [westWall, eastWall, southWall, northWall]:  # + edges:
            prims.extend(fig.primitives(camera))
        return prims




class Wall(morpho.SpaceFigure):

    def __init__(
        self, func=lambda v: v, start=morpho.array([0,0]), end=morpho.array([1,0]),
        strokeWeight=3, color=(0,0,0), alphaEdge=1, fill=(1,0,0), alphaFill=1, alpha=1,
        steps=50, origin=morpho.array([0,0,0])):

        # Construct default figure
        super().__init__()

        # Define tweenables
        func = morpho.Tweenable("func", func, tags=["function"])
        _start = morpho.Tweenable("_start", morpho.array(start), tags=["nparray"])
        _end = morpho.Tweenable("_end", morpho.array(end), tags=["nparray"])
        strokeWeight = morpho.Tweenable("strokeWeight", strokeWeight, tags=["scalar"])
        color = morpho.Tweenable("color", color, tags=["color"])
        alphaEdge = morpho.Tweenable("alphaEdge", alphaEdge, tags=["scalar"])
        alphaFill = morpho.Tweenable("alphaFill", alphaFill, tags=["scalar"])
        fill = morpho.Tweenable("fill", fill, tags=["color"])
        alpha = morpho.Tweenable("alpha", alpha, tags=["scalar"])
        steps = morpho.Tweenable("steps", steps, tags=["integer"])
        _origin = morpho.Tweenable("_origin", morpho.array(origin), tags=["nparray"])

        self.update([func, _start, _end, strokeWeight, color, fill,
            alpha, alphaEdge, alphaFill, steps, _origin])

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        self._start = morpho.array(value)

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        self._end = morpho.array(value)

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = morpho.array(value)

    def primitives(self, camera):

        # v0 = np.pad(self.start, (0,1))
        # v1 = np.pad(self.end, (0,1))
        v0 = self.start
        v1 = self.end
        top = morpho.grid.spaceline(v0, v1, steps=self.steps)

        # def f(v):
        #     x,y,z = v
        #     X = x
        #     Y = y
        #     Z = self.func([x,y])

        #     return morpho.array([X,Y,Z])

        top = top.fimage(self.func)
        vertices = top.seq + [v1, v0]
        wall = morpho.grid.SpacePolygon(vertices, width=self.strokeWeight,
            color=self.color, alphaEdge=self.alphaEdge, fill=self.fill,
            alphaFill=self.alphaFill, alpha=self.alpha)
        if not np.allclose(self.origin, (0,0,0)):
            wall = wall.fimage(lambda v: v + self.origin)

        return wall.primitives(camera)


class RiemannRect(morpho.Figure):

    def __init__(self, width=1, height=1, pos=0,
        color=(0,0,0), fill=(1,0,0), alpha=1, align=0,
        strokeWeight=3):
        # Construct a default figure
        super().__init__()

        width = morpho.Tweenable("width", width, tags=["scalar"])
        height = morpho.Tweenable("height", height, tags=["scalar"])
        pos = morpho.Tweenable("pos", pos, tags=["complex", "position"])
        color = morpho.Tweenable("color", list(color), tags=["color"])
        fill = morpho.Tweenable("fill", list(fill), tags=["color"])
        alpha = morpho.Tweenable("alpha", alpha, tags=["scalar"])
        align = morpho.Tweenable("align", align, tags=["scalar"])
        strokeWeight = morpho.Tweenable("strokeWeight", strokeWeight, tags=["scalar"])

        # Initialize tweenables
        self.update([width, height, pos, color, fill, alpha, align, strokeWeight])

        # Initialize polygon reference figure
        self.polygon = morpho.grid.Polygon(
            vertices=[0,0,0,0], width=self.strokeWeight, color=self.color,
            fill=self.fill, alphaFill=1, alpha=self.alpha
            )
        self.updatePolygon()

    # Updates polygon reference figure
    def updatePolygon(self):
        # Locate corners of the rectangle
        SW = -self.width/2*(self.align + 1) + self.pos
        NW = SW + self.height*1j
        NE = NW + self.width
        SE = SW + self.width

        self.polygon.vertices = [SW, NW, NE, SE]
        self.polygon.width = self.strokeWeight
        self.polygon.color = self.color[:]
        self.polygon.fill = self.fill[:]
        self.polygon.alpha = self.alpha

    # Again, not sure this is needed because generic copy()
    # will call the constructor thereby initializing a new
    # internal polygon.
    def copy(self):
        C = super().copy()
        C.polygon = self.polygon.copy()
        return C

    def draw(self, camera, ctx):
        self.updatePolygon()
        self.polygon.draw(camera, ctx)

class RiemannDisk(morpho.Figure):

    def __init__(self, thickness=1, radius=1, tilt=0.1, pos=0,
        outlineColor=(0,0,0), faceFill=(0.25,0.25,0.5), edgeFill=(0.5, 0.5, 1),
        alpha=1, align=0, strokeWeight=3):
        # Construct a default figure
        super().__init__()

        thickness = morpho.Tweenable("thickness", thickness, tags=["scalar"])
        radius = morpho.Tweenable("radius", radius, tags=["scalar"])
        tilt = morpho.Tweenable("tilt", tilt, tags=["scalar"])
        pos = morpho.Tweenable("pos", pos, tags=["complex", "position"])
        outlineColor = morpho.Tweenable("outlineColor", list(outlineColor), tags=["color"])
        faceFill = morpho.Tweenable("faceFill", list(faceFill), tags=["color"])
        edgeFill = morpho.Tweenable("edgeFill", list(edgeFill), tags=["color"])
        alpha = morpho.Tweenable("alpha", alpha, tags=["scalar"])
        align = morpho.Tweenable("align", align, tags=["scalar"])
        strokeWeight = morpho.Tweenable("strokeWeight", strokeWeight, tags=["scalar"])

        # Initialize tweenables
        self.update([thickness, radius, tilt, pos, outlineColor,
            faceFill, edgeFill, alpha, align, strokeWeight])

        # Initialize ellipse reference figure
        self.face = morpho.shapes.Ellipse(fillAlpha=1)
        self.updateFace()

    # Update ellipse figure representing the disk's face.
    def updateFace(self):
        self.face.pos = self.pos - (self.align-1)/2*self.thickness
        self.face.xradius = self.tilt*self.radius
        self.face.yradius = self.radius
        self.face.strokeWeight = self.strokeWeight
        self.face.color = self.outlineColor[:]
        self.face.fill = self.faceFill[:]
        self.face.alpha = self.alpha

    def draw(self, camera, ctx):
        view = camera.view

        self.updateFace()
        # if self.face.xradius == 0 or self.face.yradius == 0:
        #     return

        # Draw the disk edge first.
        # Find peak of the ellipse and move the pen to it.
        z = self.face.pos + self.face.yradius*1j
        X,Y = morpho.screenCoords(z, view, ctx)
        ctx.move_to(X,Y)

        # Draw upper edge of disk
        WIDTH = morpho.pixelWidth(self.thickness, view, ctx)
        ctx.rel_line_to(-WIDTH, 0)

        # Draw leftmost edge of disk
        ctx.save()
        X,Y = morpho.screenCoords(self.face.pos, view, ctx)
        ctx.translate(X-WIDTH, Y)
        ctx.scale(
            max(morpho.pixelWidth(self.face.xradius, view, ctx), 0.1),
            max(morpho.pixelHeight(self.face.yradius, view, ctx), 0.1)
            )
        ctx.arc(0,0, 1, deg90, deg270)
        ctx.restore()

        # Draw lower edge of disk
        ctx.rel_line_to(WIDTH, 0)

        # Close the path and fill the remainder with the edge fill
        ctx.close_path()
        ctx.set_source_rgba(*self.edgeFill, self.alpha)
        ctx.fill_preserve()

        # Stroke path
        ctx.set_source_rgba(*self.outlineColor, self.alpha)
        ctx.set_line_width(self.strokeWeight)
        ctx.stroke()

        # Now draw the face
        self.face.draw(camera, ctx)


# Constructs the Riemann rectangles for a given function
# over a given interval.
# align = -1 => left-hand rects
# align = 1 => right-hand rects
# align = 0 => midpoints
def RiemannSum(func, interval, rectCount, align=-1,
    color=(0,0,0), fill=(1, 0.3, 0.3), strokeWeight=3,
    transition=morpho.transitions.slow_fast_slow):

    a,b = interval
    dx = (b-a)/rectCount

    frame = morpho.anim.Frame()
    for i in range(rectCount):
        x = a + (i + (align+1)/2)*dx
        rect = RiemannRect(
            width=dx, height=func(x), pos=x,
            color=list(color), fill=list(fill),
            align=align, strokeWeight=strokeWeight
            )
        rect.transition = transition
        frame.figures.append(rect)

    return frame

def RiemannSumApproach(
    func, interval, initRectCount=2,
    align=-1, doublings=1,
    tweenDuration=30, pauseDuration=30,
    color=(0,0,0), fill=(1, 0.3, 0.3), strokeWeight=3,
    view=(-5,5, -5,5), transition=morpho.transitions.slow_fast_slow):

    frame0 = RiemannSum(
        func, interval, initRectCount, align, color, fill, strokeWeight, transition
        )
    frames = [frame0]

    # Construct other Riemann sums
    rectCount = initRectCount
    for n in range(doublings):
        rectCount *= 2
        frame = RiemannSum(
            func, interval, rectCount, align, color, fill, strokeWeight, transition
            )
        frames.append(frame)

    # # Test animate
    # mation = morpho.Animation()
    # for n in range(len(frames)):
    #     frame = frames[n]
    #     frame.delay = tweenDuration + pauseDuration - 1
    #     frame = morpho.Actor(frame)
    #     mation.merge(frame, atFrame=(n)*(tweenDuration+pauseDuration))
    # return mation

    # Animate!
    mation = morpho.Animation()
    for n in range(len(frames)-1):
        frame = frames[n]
        # frame.delay = pauseDuration
        # layer = morpho.anim.Layer(frame, view=view)
        # mation.merge(layer, atFrame=n*(tweenDuration))

        # Create doubled-in-place version
        copy1 = frame.copy()
        copy2 = frame.copy()
        doubledFrame = morpho.anim.Frame()
        for j in range(len(frame.figures)):
            doubledFrame.figures.append(copy1.figures[j])
            doubledFrame.figures.append(copy2.figures[j])

        # Create a film where the doubled frame turns into
        # the next frame
        film = morpho.Actor(doubledFrame)
        film.newkey(tweenDuration, frames[n+1])

        layer = morpho.Layer(film, view=view)
        mation.merge(layer, atFrame=n*tweenDuration)

    for n in range(len(frames)):
        mation.delays[n*tweenDuration] = pauseDuration

    return mation

def RiemannDiskSum(func, interval, diskCount, align=-1, tilt=0.3,
    outlineColor=(0,0,0), faceFill=(0.25,0.25,0.5), edgeFill=(0.5, 0.5, 1),
    strokeWeight=3,
    transition=morpho.transitions.slow_fast_slow):

    a,b = interval
    dx = (b-a)/diskCount

    frame = morpho.anim.Frame()
    for i in range(diskCount):
        x = a + (i + (align+1)/2)*dx
        disk = RiemannDisk(
            thickness=dx, radius=func(x), tilt=tilt, pos=x,
            outlineColor=outlineColor, faceFill=faceFill, edgeFill=edgeFill,
            align=align, strokeWeight=3
            )
        disk.transition = transition
        frame.figures.append(disk)

    return frame

def RiemannDiskSumApproach(
    func, interval, initDiskCount=2,
    align=-1, tilt=0.3, doublings=1,
    tweenDuration=30, pauseDuration=30,
    outlineColor=(0,0,0), faceFill=(0.25,0.25,0.5), edgeFill=(0.5, 0.5, 1),
    strokeWeight=3,
    view=(-5,5, -5,5), transition=morpho.transitions.slow_fast_slow):

    frame0 = RiemannDiskSum(
        func, interval, initDiskCount, align, tilt,
        outlineColor, faceFill, edgeFill, strokeWeight,
        transition
        )
    frames = [frame0]

    # Construct other Riemann sums
    diskCount = initDiskCount
    for n in range(doublings):
        diskCount *= 2
        frame = RiemannDiskSum(
            func, interval, diskCount, align, tilt,
            outlineColor, faceFill, edgeFill, strokeWeight,
            transition
            )
        frames.append(frame)

    # Animate!
    mation = morpho.Animation()
    for n in range(len(frames)-1):
        frame = frames[n]
        # frame.delay = pauseDuration
        # layer = morpho.anim.Layer(frame, view=view)
        # mation.merge(layer, atFrame=n*(tweenDuration))

        # Create doubled-in-place version
        copy1 = frame.copy()
        copy2 = frame.copy()
        doubledFrame = morpho.anim.Frame()
        for j in range(len(frame.figures)):
            doubledFrame.figures.append(copy1.figures[j])
            doubledFrame.figures.append(copy2.figures[j])

        # Create a film where the doubled frame turns into
        # the next frame
        film = morpho.Actor(doubledFrame)
        film.newkey(tweenDuration, frames[n+1])

        layer = morpho.Layer(film, view=view)
        mation.merge(layer, atFrame=n*tweenDuration)

    for n in range(len(frames)):
        mation.delays[n*tweenDuration] = pauseDuration

    return mation

# Revolves a copy of the given point by the specified angle.
def revolvePoint(point, theta, axis="x", axisOffset=0, stretch=0.2):
    point = point.copy()
    if axis.lower() == "x":
        C0 = point.pos.real + axisOffset*1j  # Center of rotation
        point.pos = (point.pos - C0)*exp(theta*1j)
        point.pos = stretch*point.pos.real + point.pos.imag*1j
        point.pos += C0
    elif axis.lower() == "y":
        pass
    else:
        raise ValueError('axis must be "x" or "y"')
    return point

# Revolves a copy of the given path by the specified angle.
def revolvePath(path, theta, axis="x", axisOffset=0, tilt=0.1):
    path = path.copy()
    if axis.lower() == "x":
        for n in range(len(path.seq)):
            z = path.seq[n]
            C0 = z.real + axisOffset*1j  # Center of rotation
            z = (z - C0)*exp(theta*1j)
            z = tilt*z.real + z.imag*1j
            z += C0

            path.seq[n] = z
    elif axis.lower() == "y":
        pass
    else:
        raise ValueError('axis must be "x" or "y"')
    return path

def revolvePathSolid(path, theta,
    axis="x", axisOffset=0, tilt=0.1, dtheta=0.03,
    angleSpacing=pi/6, Nsubarcs=10,
    fill=(0.35, 0.35, 1), innerfill=(0.35/2, 0.35/2, 1/2)):

    path = path.copy()
    if axis.lower() == "x":
        # Constrain theta between 0 and tau
        theta = min(max(theta, 0), tau)

        pathFinal = revolvePath(path, min(theta, pi), axis, axisOffset, tilt)

        # Make first-pass righthand boundary vertices
        rverts1 = []
        pt = morpho.grid.Point(path.seq[-1])
        for n in range(1, int(min(theta, pi)//dtheta)):
            rverts1.append(revolvePoint(pt, n*dtheta, axis, axisOffset, tilt).pos)
        if theta >= pi:
            rverts1.append(revolvePoint(pt, pi, axis, axisOffset, tilt).pos)

        # Make second-pass righthand boundary vertices
        rverts2 = []
        for n in range(1, int((theta-pi)//dtheta)):
            rverts2.append(revolvePoint(pt, pi+n*dtheta, axis, axisOffset, tilt).pos)
        if theta == tau:
            rverts2.append(pt.pos)

        # Make lefthand boundary vertices
        lverts = []
        pt = morpho.grid.Point(path.seq[0])
        for n in range(1, int(min(theta, pi)//dtheta)):
            lverts.append(revolvePoint(pt, n*dtheta, axis, axisOffset, tilt).pos)
        if theta >= pi:
            lverts.append(revolvePoint(pt, pi, axis, axisOffset, tilt).pos)

        # Define main polygon
        mainpoly = morpho.grid.Polygon(
            vertices=path.seq[:] + rverts1[:] + pathFinal.seq[::-1] + lverts[::-1],
            width=path.width, color=path.color[:], fill=fill,
            alphaFill=1
            )

        frame = morpho.Frame([mainpoly])

        # Define inner polygon
        if theta > pi:
            pathFinalTrue = revolvePath(
                path, theta, axis, axisOffset, tilt
                )
            innerpoly = morpho.grid.Polygon(
                vertices=rverts2[:] + [path.seq[0].real + 1j*rverts2[-1].imag, lverts[-1]],
                width=path.width, color=path.color[:],
                fill=innerfill, alphaFill=1
                )
            frame.figures.insert(0, innerpoly)

        # Define subpaths
        subpaths = []
        for n in range(1, int(min(theta, pi)//angleSpacing + 1)):
            angle = n*angleSpacing
            # Don't render a subpath at pi.
            if abs(angle-pi) < 1e-9:
                continue
            subpath = revolvePath(
                path, angle, axis, axisOffset, tilt
                )
            subpath.width /= 2
            subpaths.append(subpath)
        frame.figures.extend(subpaths)

        # Define subarcs
        subarcs = []
        len_seq = len(path.seq)
        for n in range(1, Nsubarcs+1):
            index = len_seq*n//(Nsubarcs+1)
            pos = path.seq[index].real + axisOffset*1j
            yradius = path.seq[index].imag - axisOffset
            xradius = tilt*yradius
            arc = morpho.shapes.EllipticalArc(
                pos, xradius, yradius,
                theta0=tau/4, theta1=tau/4+min(theta, pi),
                strokeWeight=path.width/2, color=path.color
                )
            subarcs.append(arc)
        frame.figures.extend(subarcs)

        return frame

    elif axis.lower() == "y":
        pass
    else:
        raise ValueError('axis must be "x" or "y"')



'''
CREDIT: Thanks to Mr. B's answer on StackOverflow
https://stackoverflow.com/a/25562415
'''

import importlib
import pkgutil


def import_submodules(package, recursive=True):
    """ Import all submodules of a module, recursively, including subpackages

    :param package: package (name or actual module)
    :type package: str | module
    :rtype: dict[str, types.ModuleType]
    """
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + '.' + name
        results[full_name] = importlib.import_module(full_name)
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results



# Provides a class of stopwatch objects.
import time

'''
Stopwatch object. Measures the duration between a .tic() call
and a .toc() call in units of seconds.'''
class Timer(object):
    # Constructor. Initializes startTime to
    # the time of construction.
    def __init__(self):
        self.startTime = time.time()

    # Resets the timer.
    def tic(self):
        self.startTime = time.time()

    # Returns time elapsed (in seconds) since the last
    # .tic() was called.
    def toc(self):
        return time.time() - self.startTime

# Function waits for the condition specified by cond to be True.
# cond is an expression given using lambda.
# e.g. waitfor(lambda: fileExists("someFile.txt"))
# "until" is the maximum amount of time to wait for cond to be True
# Defaults to infinity.
# "period" is how many seconds to wait before checking again if cond is True
# Defaults to 1 second.
# waitfor() returns True if cond becomes True
# waitfor() returns False if until exceeded, and so waitfor() gave up.
# NOTE: cond is evaluated lazily. waitfor() uses the variable names:
# cond, __waitfor_timer, until, time, period
# If you use the same names in your cond, waitfor() may act badly.
# Hopefully, your expression for cond doesn't use any of these names,
# but just to be safe, the correct way to provide input into cond is to
# explicitly declare your variables like so:
#
# lambda x=x, y=y, z=z: your_expression_containing_xyz
#
# That way, when cond is evaluated, x, y, and z will be YOUR x, y, and z.
def waitfor(cond, until=float("inf"), period=1):
    if type(cond) is str:
        cond = lambda cond=cond: eval(cond)
    __waitfor_timer = Timer()
    while not cond() and __waitfor_timer.toc() < until:
        time.sleep(period)
    return cond()

# Object used for the tic() and toc() functions that can
# be used in the global scope.
masterTimer = Timer()

# Resets the master timer
def tic():
    masterTimer.tic()

# Returns seconds elapsed since last tic() was called.
def toc():
    return masterTimer.toc()



'''
This module contains functions to facilitate exporting animations
as animated GIFs.
'''

import imageio
import os, platform
import subprocess as sp

# Code tells sp.call() not to make a console window (for Windows)
CREATE_NO_WINDOW = 0x08000000

# # Get location of gifsicle
# pwd = os.path.dirname(os.path.abspath(__file__)) + os.sep # + "gifsicle"
# pwd += os.sep

pwd = os.curdir + os.sep

# Compiles the set of image files into an animated gif.
# INPUTS:
# filenames = List of filenames (or file paths)
# directory = Directory of the filenames (defaults to "./")
# saveas    = filename of the exported gif
# duration  = Duration of each frame in seconds.
#             Can be a single number to apply to all frames,
#             or a list denoting the duration of each filename.
#             Defaults to 0.1 seconds.
#             Maximum: 655 seconds. If any duration exceeds this value,
#             it will be lowered to the maximum.
def makegif(filenames="*", directory=pwd, saveas=pwd+"movie.gif", \
    duration=0.1):
    # Lower overflows
    if type(duration) is int or type(duration) is float:
        duration = min(duration, 655)
    else:
        for i in range(len(duration)):
            duration[i] = min(duration[i], 655)

    if filenames == "*": filenames = os.listdir(directory)
    initializegif(saveas)
    with imageio.get_writer(saveas, mode='I', duration=duration) as writer:
        for filename in filenames:
            image = imageio.imread(directory + os.sep + filename)
            writer.append_data(image)

# Optimizes a gif file in place using gifsicle.
# Requires the gifsicle executable to be in the current directory.
def optimizegif(filename):
    if not os.path.isfile(filename):
        raise FileNotFoundError
    if platform.system() == "Windows":
        cmd = ["gifsicle", "-b", "-O3", "--careful", filename]
        sp.call(cmd, creationflags=CREATE_NO_WINDOW)
        # sp.call(cmd)
    else:
        cmd = ["gifsicle", "-b", "-O3", "--careful", filename]
        sp.call(cmd)

# Attempts to initialize an empty GIF file.
# If file already exists, does nothing.
# Although this is not technically necessary for GIF creation,
# it allows a way to test a filepath's validity.
def initializegif(filename):
    # if filename exists, then nothing needs to be done.
    if os.path.isfile(filename):
        return
    # Try to make the file.
    with open(filename, "a") as file:
        pass



'''
This module contains some helper functions that automate some common
animation actions.
'''

import morpholib as morpho
# import morpho, morpho.anim
# import morpho.grid
# from morpho.tools.basics import *
# import math, cmath

# TODO: Add "stagger" argument to both fadeOut() and rollback()

def fadeOut(actors, duration=30, atFrame=None):
    # if not isinstance(actors, list) and not isinstance(actors, tuple):
    #     actors = [actors]
    # Turn into a list if necessary
    if isinstance(actors, morpho.Actor):
        actors = [actors]
    if atFrame is None:
        atFrame = max(actor.lastID() for actor in actors)
    for actor in actors:
        actor.newkey(atFrame)
        actor.newendkey(duration).alpha = 0
        actor.last().visible = False

def rollback(actors, duration=30, atFrame=None):
    # if not isinstance(actors, list) and not isinstance(actors, tuple):
    #     actors = [actors]
    # Turn into a list if necessary
    if isinstance(actors, morpho.Actor):
        actors = [actors]
    if atFrame is None:
        atFrame = max(actor.lastID() for actor in actors)
    for actor in actors:
        actor.newkey(atFrame)
        actor.newendkey(duration, actor.first().copy())
        actor.last().visible = False

# NOT IMPLEMENTED!
# Transforms one figure into another using a fade effect.
# Requires the underlying figures to have "pos" and "alpha"
# attributes to work.
# Returns an animation in which fig move toward pig while
# fading into pig.
def transform(fig, pig, time=30):
    raise NotImplementedError
